# -----------------------------------------------------------------------------
# Copyright (c) 2013-2025, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------
import inspect
import re
from itertools import cycle

import numpy as np
from lmfit import Model, Parameters
from lmfit import __version__ as lmfit_version
from nexusformat.nexus import (NeXusError, NXdata, NXentry, NXfield, NXnote,
                               NXparameters, NXprocess, nxload)

from .plotview import NXPlotView, linestyles
from .pyqt import QtCore, QtWidgets
from .utils import display_message, format_float, load_models, report_error
from .widgets import (NXCheckBox, NXColorBox, NXComboBox, NXDialog, NXLabel,
                      NXLineEdit, NXMessageBox, NXPanel, NXPushButton,
                      NXrectangle, NXScrollArea, NXTab)


def get_models():
    """
    Return a dictionary of LMFIT models.

    This function returns a dictionary of LMFIT models, including those
    defined in the LMFIT package and those defined in the
    ``nexpy.models`` package. Additional models can also be defined in
    the ``~/.nexpy/models`` directory or in another installed package,
    which declares the entry point ``nexpy.models``. The models are
    returned as a dictionary where the keys are the names of the models
    and the values are the classes defining the models.
    """
    from lmfit.models import lmfit_models
    models = lmfit_models
    if 'Expression' in models:
        del models['Expression']
    if 'Gaussian-2D' in models:
        del models['Gaussian-2D']

    nexpy_models = load_models()

    for model in nexpy_models:
        try:
            models.update(
                dict((n.strip('Model'), m)
                for n, m in inspect.getmembers(nexpy_models[model],
                                               inspect.isclass)
                if issubclass(m, Model) and n != 'Model'))
        except ImportError:
            pass

    return models


all_models = get_models()


def get_methods():
    """Return a dictionary of minimization methods in LMFIT."""
    methods = {'leastsq': 'Levenberg-Marquardt',
               'least_squares': 'Least-Squares minimization, '
                                'using Trust Region Reflective method',
               'differential_evolution': 'differential evolution',
               'nelder': 'Nelder-Mead',
               'lbfgsb': ' L-BFGS-B',
               'powell': 'Powell',
               'cg': 'Conjugate-Gradient',
               'newton': 'Newton-CG',
               'cobyla': 'Cobyla',
               'bfgs': 'BFGS',
               'tnc': 'Truncated Newton',
               'trust-ncg': 'Newton-CG trust-region',
               'trust-exact': 'nearly exact trust-region',
               'trust-krylov': 'Newton GLTR trust-region',
               'trust-constr': 'trust-region for constrained optimization',
               'dogleg': 'Dog-leg trust-region',
               'slsqp': 'Sequential Linear Squares Programming'}
    return methods


all_methods = get_methods()


class FitDialog(NXPanel):

    def __init__(self, parent=None):
        """
        Initialize the Fit Panel.

        Parameters
        ----------
        parent : QWidget, optional
            The parent of the dialog. The default is None.
        """
        super().__init__('Fit', title='Fit Panel', apply=True, reset=True,
                         parent=parent)
        self.setMinimumWidth(850)
        self.tab_class = FitTab

    def activate(self, data, plotview=None, color='C0'):
        """
        Activate the Fit Panel.

        Parameters
        ----------
        data : NXdata
            data to fit
        plotview : NXPlotView, optional
            The plotview containing the data to fit. The default is None.
        color : str, optional
            The color of the fit curve. The default is 'C0'.
        """
        
        if plotview:
            label = plotview.label + ': ' + str(plotview.num)
        else:
            label = data.nxroot.nxname + data.nxpath
        super().activate(label, data, plotview=plotview, color=color)


class FitTab(NXTab):

    def __init__(self, label, data, plotview=None, color='C0', parent=None):

        """
        Initialize the Fit Tab.

        Parameters
        ----------
        label : str
            The label for the tab.
        data : NXdata
            The data to fit.
        plotview : NXPlotView, optional
            The plotview containing the data to fit. The default is None.
        color : str, optional
            The color of the fit curve. The default is 'C0'.
        parent : QWidget, optional
            The parent of the tab. The default is None.
        """
        super().__init__(label, parent=parent)

        if ((isinstance(data, NXentry) or isinstance(data, NXprocess))
                and 'data' in data):
            group = data
            self.initialize_data(group['data'])
        elif isinstance(data, NXdata):
            self.initialize_data(data)
            group = None
        else:
            raise NeXusError("Must be an NXdata group")

        self.plotview = plotview
        self.plot_nums = []

        self.model = None
        self.models = []

        self.first_time = True
        self.fitted = False
        self.fit = None

        self.initialize_models()

        add_button = NXPushButton("Add Model", self.add_model)
        self.model_combo = NXComboBox(items=list(self.all_models),
                                      slot=self.choose_model)
        if 'Gaussian' in self.model_combo:
            self.model_combo.select('Gaussian')
        try:
            from pylatexenc.latex2text import LatexNodes2Text
            text = LatexNodes2Text().latex_to_text
        except ImportError:
            text = str
        for i, m in enumerate(self.all_models):
            tooltip = self.all_models[m].__doc__
            if tooltip:
                tooltip = tooltip.replace('.. math::\n\n', '')
                tooltip = re.sub(r'\:[a-z]*\:', r'', tooltip)
                self.model_combo.setItemData(i, text(tooltip),
                                             QtCore.Qt.ToolTipRole)
        self.form_combo = NXComboBox()
        self.compose_button = NXPushButton(
            "Compose Models", self.compose_model)
        self.import_button = NXPushButton("Import Model", self.import_fit)
        model_layout = self.make_layout(add_button, self.model_combo,
                                        self.form_combo, 'stretch',
                                        self.import_button,
                                        self.compose_button, align='justified')

        self.parameter_layout = self.initialize_parameter_grid()

        self.remove_button = NXPushButton("Remove Model", self.remove_model)
        self.remove_combo = NXComboBox()
        self.restore_button = NXPushButton("Restore Parameters",
                                           self.restore_parameters)
        self.copy_parameters_button = NXPushButton("Copy Parameters",
                                                   self.copy_parameters)
        self.remove_layout = self.make_layout(self.remove_button,
                                              self.remove_combo,
                                              'stretch',
                                              self.restore_button,
                                              self.copy_parameters_button,
                                              align='justified')

        if self.plotview is None:
            self.plotview = NXPlotView('Fit')
            self.plotview.plot(self._data, fmt='o', color=color)
        self.data_num = self.plotview.num
        self.data_label = self.plotview.plots[self.plotview.num]['label']
        self.cursor = self.plotview.plots[self.plotview.num]['cursor']
        if self.cursor:
            @self.cursor.connect("add")
            def add_selection(sel):
                self.mask_data()

        self.fit_button = NXPushButton('Fit', self.fit_data)
        self.fit_combo = NXComboBox(items=list(all_methods))
        for i, m in enumerate(all_methods):
            tooltip = all_methods[m]
            if tooltip:
                self.fit_combo.setItemData(i, text(tooltip),
                                           QtCore.Qt.ToolTipRole)
        self.fit_combo.sort()
        self.fit_combo.select('leastsq')
        if self._data.nxerrors:
            self.fit_checkbox = NXCheckBox('Use Errors', checked=True)
        else:
            self.fit_checkbox = NXCheckBox('Use Poisson Errors',
                                           self.define_errors)
        self.copy_fit_button = NXPushButton("Copy Fit", self.copy_fit)
        self.save_fit_button = NXPushButton("Save Fit", self.save_fit)
        self.adjust_layout = self.make_layout(self.fit_button,
                                              self.fit_combo,
                                              self.fit_checkbox,
                                              'stretch',
                                              self.copy_fit_button,
                                              self.save_fit_button,
                                              align='justified')

        self.fit_status = NXLabel(width=600)
        self.report_button = NXPushButton("Show Fit Report", self.report_fit)
        self.action_layout = self.make_layout(self.fit_status,
                                              'stretch',
                                              self.report_button,
                                              align='justified')

        plot_data_button = NXPushButton('Plot Data', self.plot_data)
        self.plot_model_button = NXPushButton('Plot Model', self.plot_model)
        self.plot_combo = NXComboBox()
        self.plot_checkbox = NXCheckBox('Use Data Points')
        self.mask_button = NXPushButton('Mask Data', self.mask_data)
        self.clear_mask_button = NXPushButton('Clear Masks', self.clear_masks)
        self.color_box = NXColorBox(color, label='Plot Color', width=100)
        self.plot_layout = self.make_layout(plot_data_button,
                                            self.plot_model_button,
                                            self.plot_combo,
                                            self.plot_checkbox,
                                            self.mask_button,
                                            self.clear_mask_button,
                                            'stretch',
                                            self.color_box,
                                            align='justified')
        self.clear_mask_button.setVisible(False)

        self.set_layout(model_layout, self.plot_layout)
        self.layout.setSpacing(5)
        self.set_title("Fit NeXus Data")
        self.choose_model()
        self.set_button_visibility()

        self.cid = self.plotview.canvas.mpl_connect('button_release_event',
                                                   self.on_button_release)
        self.composite_model = ''
        self.composite_dialog = None
        self.plot_dialog = None
        self.expression_dialog = None
        self.rectangle = None
        self.mask_num = None
        self.linestyles = [linestyles[ls] for ls in linestyles
                           if ls != 'Solid' and ls != 'None']
        self.linestyle = cycle(self.linestyles)
        self.xlo, self.xhi = self.plotview.ax.get_xlim()
        self.ylo, self.yhi = self.plotview.ax.get_ylim()

        if group:
            self.load_fit(group)

    def __repr__(self):
        return f'FitTab("{self.data_label}")'

    @property
    def fitview(self):
        """The plotting window for the fitting data."""
        if self.plotview and self.plotview.label in self.plotviews:
            return self.plotview
        elif 'Fit' in self.plotviews:
            self.plotview = self.plotviews['Fit']
            return self.plotview
        else:
            return None

    @fitview.setter
    def fitview(self, plotview):
        self.plotview = plotview

    def initialize_data(self, data):
        """
        Initialize the data to be fitted.

        Parameters
        ----------
        data : NXdata
            The data to be fitted.

        Raises
        ------
        NeXusError
            If the data is not one-dimensional.
        """
        if isinstance(data, NXdata):
            if len(data.shape) > 1:
                raise NeXusError(
                    "Fitting only possible on one-dimensional arrays")
            self._data = NXdata()
            self._data['signal'] = data.nxsignal
            self._data.nxsignal = self._data['signal']
            if data.nxaxes[0].size == data.nxsignal.size + 1:
                self._data['axis'] = data.nxaxes[0].centers()
            elif data.nxaxes[0].size == data.nxsignal.size:
                self._data['axis'] = data.nxaxes[0]
            else:
                raise NeXusError("Data has invalid axes")
            self._data.nxaxes = [self._data['axis']]
            if data.nxerrors:
                self._data.nxerrors = data.nxerrors
                self.poisson_errors = False
            else:
                self.poisson_errors = True
            self._data['title'] = data.nxtitle
        else:
            raise NeXusError("Must be an NXdata group")

    def initialize_models(self):
        """Initialize the list of models."""
        self.all_models = all_models

    def initialize_parameter_grid(self):
        """
        Initialize the parameter grid layout.

        The parameter grid is a table of parameters for each model
        with columns for the model name, parameter name, value, min,
        max, and whether the parameter is fixed or not. The parameter
        grid is located in a scroll area with a minimum height and
        width to prevent it from being too small.

        Returns
        -------
        grid_layout : QGridLayout
            The layout of the parameter grid.
        """
        grid_layout = QtWidgets.QVBoxLayout()

        self.parameter_grid = QtWidgets.QGridLayout()
        self.parameter_grid.setSpacing(5)
        headers = ['Model', 'Name', 'Value', '', 'Min', 'Max', 'Fixed', '']
        width = [100, 100, 100, 100, 100, 100, 50, 50]
        column = 0
        for header in headers:
            label = NXLabel(header, bold=True, align='center')
            self.parameter_grid.addWidget(label, 0, column)
            self.parameter_grid.setColumnMinimumWidth(column, width[column])
            column += 1

        self.scroll_widget = QtWidgets.QWidget()
        self.scroll_area = NXScrollArea(self.scroll_widget)
        self.scroll_layout = QtWidgets.QVBoxLayout()
        self.scroll_layout.addLayout(self.parameter_grid)
        self.scroll_layout.addStretch()
        self.scroll_widget.setLayout(self.scroll_layout)
        self.scroll_area.setMinimumHeight(200)
        self.scroll_area.setMinimumWidth(800)
        self.scroll_area.setSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                       QtWidgets.QSizePolicy.Expanding)

        grid_layout.addWidget(self.scroll_area)
        return grid_layout

    def set_button_visibility(self, fitted=False):
        """
        Set the visibility of the buttons in the dialog based on whether
        there are any models defined and whether the data have been
        fitted.

        Parameters
        ----------
        fitted : bool, optional
            Set to True if the data have been fitted.
        """
        if len(self.models) == 0:
            self.import_button.setVisible(True)
            self.compose_button.setVisible(False)
            self.remove_button.setVisible(False)
            self.remove_combo.setVisible(False)
            self.restore_button.setVisible(False)
            self.copy_parameters_button.setVisible(False)
            self.fit_button.setVisible(False)
            self.fit_combo.setVisible(False)
            self.fit_checkbox.setVisible(False)
            self.fit_status.setVisible(False)
            self.report_button.setVisible(False)
            self.copy_fit_button.setVisible(False)
            self.save_fit_button.setVisible(False)
            self.plot_model_button.setVisible(False)
            self.plot_combo.setVisible(False)
            self.plot_checkbox.setVisible(False)
        else:
            self.import_button.setVisible(False)
            if len(self.models) > 1:
                self.compose_button.setVisible(True)
            self.remove_button.setVisible(True)
            self.remove_combo.setVisible(True)
            self.copy_parameters_button.setVisible(True)
            self.fit_button.setVisible(True)
            self.fit_combo.setVisible(True)
            self.fit_checkbox.setVisible(True)
            self.plot_model_button.setVisible(True)
            self.plot_combo.setVisible(True)
            self.plot_checkbox.setVisible(True)
            if fitted:
                self.restore_button.setVisible(True)
                self.fit_status.setVisible(True)
                self.report_button.setVisible(True)
                self.copy_fit_button.setVisible(True)
                self.save_fit_button.setVisible(True)
            else:
                self.copy_fit_button.setVisible(False)
                self.save_fit_button.setVisible(False)

    @property
    def data(self):
        """The data to be fitted."""
        try:
            xmin, xmax = self.get_limits()
            axis = self._data.nxaxes[0]
            if xmin > axis.max() or xmax < axis.min():
                raise NeXusError('Invalid data range')
            else:
                return self._data[xmin:xmax]
        except NeXusError as error:
            report_error("Fitting data", error)

    @property
    def signal(self):
        """
        The data to be fitted as a one-dimensional array.

        If the data is masked, the mask is removed before returning the
        data.
        """
        signal = self.data['signal']
        if signal.mask:
            return signal.nxdata.compressed().astype(np.float64)
        else:
            return signal.nxdata.astype(np.float64)

    @property
    def axis(self):
        """
        The x-axis values of the data to be fitted.

        If the data is masked, the mask is removed before returning the
        axis values.
        """
        data = self.data
        signal = data['signal'].nxdata
        axis = data['axis'].nxdata.astype(np.float64)
        if isinstance(signal, np.ma.MaskedArray):
            return np.ma.masked_array(axis, mask=signal.mask).compressed()
        else:
            return axis

    @property
    def errors(self):
        """
        The data errors as a one-dimensional array.

        If the data is masked, the mask is removed before returning the
        errors. If the data has no errors, returns None.
        """
        data = self.data
        if data.nxerrors:
            errors = data.nxerrors.nxdata.astype(np.float64)
            signal = data['signal'].nxdata
            if isinstance(signal, np.ma.MaskedArray):
                return np.ma.masked_array(
                    errors, mask=signal.mask).compressed()
            else:
                return errors
        else:
            return None

    @property
    def weights(self):
        """
        The data weights as a one-dimensional array.

        If the data is masked, the mask is removed before returning the
        weights. If the data has no errors, returns None.
        """
        if self.errors is not None and np.all(self.errors):
            return 1.0 / self.errors
        else:
            return None

    def signal_mask(self):
        """
        Return the masked data as a new NXdata group.

        Returns
        -------
        NXdata
            The masked data as a new NXdata group with the same name as
            the original data, but with '_mask' appended to the name. If
            no data are masked, returns None.
        """
        mask = self._data['signal'].mask
        if mask and mask.any():
            mask_data = NXfield(self._data['signal'].nxdata.data[mask == 1],
                                name='mask')
            mask_axis = NXfield(self._data['axis'].nxdata[mask == 1],
                                name='axis')
            return NXdata(mask_data, mask_axis)
        else:
            return None

    def define_errors(self):
        """
        Define the errors for the fit.

        If the 'fit' checkbox is checked, the errors are set to the square
        root of the signal, but with a minimum of 1. If the checkbox is
        not checked, the errors attribute is deleted.
        """
        if self.poisson_errors:
            if self.fit_checkbox.isChecked():
                self._data.nxerrors = np.sqrt(np.where(self._data.nxsignal < 1,
                                                       1, self._data.nxsignal))
            else:
                del self._data[self._data.nxerrors.nxname]

    def mask_data(self):
        """
        Mask the data.

        This function can be used in two ways to mask the data:

        1. Select data with right-click zoom and click 'Mask Data'
        2. Double-click points to be masked

        The function checks if any data has been selected with the zoom tool.
        If data is selected, it is masked. Otherwise, it checks if a rectangle
        has been drawn. If a rectangle has been drawn, the data within the
        rectangle is masked. If no data is selected and no rectangle has been
        drawn, a message is displayed with instructions on how to mask the
        data.
        """
        axis = self._data['axis']
        signal = self._data['signal']
        if self.cursor and self.cursor.selections:
            try:
                idx = self.cursor.selections[0].index
            except AttributeError:
                raise NeXusError("Please update the 'mplcursors' package")
            idx = self.cursor.selections[0].index
            if np.ma.is_masked(signal.nxvalue[idx]):
                signal[idx] = np.ma.nomask
            else:
                signal[idx] = np.ma.masked
            self.cursor.remove_selection(self.cursor.selections[0])
        elif self.rectangle:
            signal[(axis >= self.xlo) & (axis <= self.xhi)] = np.ma.masked
        else:
            display_message(
                "Masking Data",
                "There are two methods to mask data:\n\n"
                "1) Select data with right-click zoom and click 'Mask Data'\n"
                "2) Double-click points to be masked", width=350)
            return
        self.plot_mask()
        if np.ma.is_masked(signal.nxvalue):
            self.mask_button.setVisible(False)
            self.clear_mask_button.setVisible(True)
        else:
            self.mask_button.setVisible(True)
            self.clear_mask_button.setVisible(False)

    def clear_masks(self):
        """
        Clear all masks from the signal.

        This function removes all masks from the signal and changes the
        visibility of the 'Mask Data' and 'Clear Masks' buttons so that
        'Mask Data' is visible and 'Clear Masks' is not.
        """
        self._data['signal'].mask = np.ma.nomask
        self.remove_masks()
        self.mask_button.setVisible(True)
        self.clear_mask_button.setVisible(False)

    @property
    def parameters(self):
        """
        The LMFIT Parameters object containing all the models.

        This property is a read-only property that returns a Parameters
        object that is a composite of all the parameters from all the
        models currently in the FitTab. It is used by the FitTab to pass
        the parameters to the LMFIT minimizer.

        Returns
        -------
        Parameters
            The LMFIT Parameters object that contains all the parameters
            for all the models.
        """
        _parameters = Parameters()
        for m in self.models:
            _parameters += m['parameters']
        return _parameters

    @parameters.setter
    def parameters(self, new_parameters):
        for m in self.models:
            for p in m['parameters']:
                if p in new_parameters:
                    m['parameters'][p].value = new_parameters[p].value
                    m['parameters'][p].min = new_parameters[p].min
                    m['parameters'][p].max = new_parameters[p].max
                    m['parameters'][p].vary = new_parameters[p].vary
                    m['parameters'][p].expr = new_parameters[p].expr
                    m['parameters'][p].stderr = new_parameters[p].stderr
                    m['parameters'][p].correl = new_parameters[p].correl

    @property
    def method(self):
        """The minimization method selected in the FitTab."""        
        return self.fit_combo.selected

    @property
    def color(self):
        """The current color as a string."""
        return self.color_box.textbox.text()

    def compressed_name(self, name):
        """Converts a model name to a compressed name"""
        return re.sub(r'([a-zA-Z_ ]*) [#] (\d*)$', r'\1_\2',
                      name, count=1).replace(' ', '_')

    def expanded_name(self, name):
        """Converts a compressed name to a model name"""
        return re.sub(r'([a-zA-Z_]*)_(\d*)$', r'\1 # \2',
                      name, count=1).replace('_', ' ').strip()

    def parse_model_name(self, name):
        """
        Parse a model name.

        The model name is expected to be of the form <model_name>_<number>,
        where <model_name> is the name of the model and <number> is the
        number of the model. The _ is optional. The function returns a tuple
        (name, number), where name is the name of the model and number is the
        number of the model. If the model name does not match the expected
        format, the function returns (None, None).

        Parameters
        ----------
        name : str
            The model name to be parsed.

        Returns
        -------
        name : str
            The name of the model.
        number : str
            The number of the model.
        """
        match = re.match(r'([a-zA-Z0-9_-]*)_(\d*)$', name)
        if match:
            return match.group(1).replace('_', ' '), match.group(2)
        try:
            match = re.match(r'([a-zA-Z]*)(\d*)', name)
            return match.group(1), match.group(2)
        except Exception:
            return None, None

    def import_fit(self):
        """Select the NXprocess group containing the fit."""
        group = self.treeview.node
        if not isinstance(group, NXprocess):
            display_message("Import Fit",
                "Select a valid NXprocess group in the tree")
        else:
            self.load_fit(group)

    def load_fit(self, group):
        """
        Load a fit from a NeXus NXprocess group.

        Parameters
        ----------
        group : NXprocess
            The NeXus NXprocess group containing the fit.
        """
        self.model = None
        self.models = []
        if 'fit' in group.entries or 'model' in group.entries:
            for name in group.entries:
                if ('parameters' in group[name] and
                        'model' in group[name]['parameters'].attrs):
                    model_class = group[name]['parameters'].attrs['model']
                    model_name = name
                else:
                    model_class, model_index = self.parse_model_name(name)
                    model_name = model_class + '_' + model_index
                    if (model_class and model_class not in self.all_models and
                            model_class+'Model' in self.all_models):
                        model_class = model_class + 'Model'
                if model_class in self.all_models:
                    model = self.get_model_instance(model_class, model_name)
                    parameters = model.make_params()
                    saved_parameters = group[name]['parameters']
                    for mp in parameters:
                        p = mp.replace(model.prefix, '')
                        p = self.convert_parameter_name(p, saved_parameters)
                        if p in saved_parameters:
                            parameter = parameters[mp]
                            parameter.value = saved_parameters[p].nxvalue
                            parameter.min = float(
                                saved_parameters[p].attrs['min'])
                            parameter.max = float(
                                saved_parameters[p].attrs['max'])
                            if 'vary' in saved_parameters[p].attrs:
                                parameter.vary = (
                                    saved_parameters[p].attrs['vary'])
                            if 'expr' in saved_parameters[p].attrs:
                                parameter.expr = (
                                    saved_parameters[p].attrs['expr'])
                            else:
                                parameter.expr = None
                            if 'error' in saved_parameters[p].attrs:
                                error = saved_parameters[p].attrs['error']
                                if error:
                                    parameter.stderr = float(
                                        saved_parameters[p].attrs['error'])
                    self.models.append({'name': model_name,
                                        'class': model_class,
                                        'model': model,
                                        'parameters': parameters})
            self.parameters = self.parameters

            def idx(model):
                return int(re.match('.*?([0-9]+)$', model['name']).group(1))
            self.models = sorted(self.models, key=idx)
            for model_index, model in enumerate(self.models):
                if model_index == 0:
                    self.model = model['model']
                    self.composite_model = model['name']
                else:
                    self.model += model['model']
                    self.composite_model += '+' + model['name']
                self.add_model_parameters(model_index)
            if 'composite_model' in group.attrs:
                composite_model = group.attrs['composite_model']
            elif 'model' in group and isinstance(group['model'], NXfield):
                composite_model = group['model'].nxvalue
            else:
                composite_model = None
            if composite_model is not None:
                self.model = self.eval_model(composite_model)
                self.composite_model = composite_model
            else:
                self.model = self.eval_model(self.composite_model)
            self.write_parameters()
            self.copy_parameters_button.setVisible(True)
            self.save_fit_button.setVisible(False)

    def convert_parameter_name(self, parameter, saved_parameters):
        """
        Convert a parameter name in the Fit Dialog to its equivalent in
        the saved parameters.

        This function is used to convert parameter names in the Fit
        Dialog to their equivalent names in the saved parameters. The
        conversion is done in the following order: 1. If the parameter
        is in the saved parameters, it is returned unchanged. 2. If the
        capitalized parameter is in the saved parameters, it is returned
           with the first letter capitalized.
        3. If the parameter is 'amplitude' and 'Integral' is in the
           saved parameters, 'Integral' is returned.
        4. If the parameter is 'sigma' and 'Gamma' is in the saved
           parameters, 'Gamma' is returned.
        5. If the parameter is 'intercept' and 'Constant' is in the
           saved parameters, 'Constant' is returned.
        6. Otherwise, an empty string is returned.

        Parameters
        ----------
        parameter : str
            The parameter name in the Fit Dialog.
        saved_parameters : dict
            The saved parameters.

        Returns
        -------
        str
            The converted parameter name.
        """
        if parameter in saved_parameters:
            return parameter
        elif parameter.capitalize() in saved_parameters:
            return parameter.capitalize()
        elif parameter == 'amplitude' and 'Integral' in saved_parameters:
            return 'Integral'
        elif parameter == 'sigma' and 'Gamma' in saved_parameters:
            return 'Gamma'
        elif parameter == 'intercept' and 'Constant' in saved_parameters:
            return 'Constant'
        else:
            return ''

    def get_model_instance(self, model_class, model_name):
        """
        Returns an instance of the specified model class.

        Parameters
        ----------
        model_class : str
            The name of the model class.
        model_name : str
            The name of the model.

        Returns
        -------
        model : Model
            An instance of the specified model class.
        """
        if self.all_models[model_class].valid_forms:
            return self.all_models[model_class](prefix=model_name+'_',
                                                form=self.form_combo.selected)
        else:
            return self.all_models[model_class](prefix=model_name+'_')

    def choose_model(self):
        """
        Slot to be called when the user selects a model class from the
        model class combo box. This slot makes the form combo box
        visible and populates it with the valid forms of the selected
        model class. If the selected model class does not have valid
        forms, the form combo box is made invisible.
        """  
        model_class = self.model_combo.selected
        try:
            if self.all_models[model_class].valid_forms:
                self.form_combo.setVisible(True)
                self.form_combo.clear()
                self.form_combo.add(*self.all_models[model_class].valid_forms)
            else:
                self.form_combo.setVisible(False)
        except AttributeError:
            self.form_combo.setVisible(False)

    def add_model(self):
        """
        Slot to be called when the user clicks the "Add" button.

        This slot adds the selected model class to the list of models,
        creates an instance of the model class, and adds its parameters
        to the model parameters box. It guesses the initial values of
        the model parameters by calling the guess method of the model
        class. If the model class does not have a guess method, it
        creates a set of parameters with default values. It then updates
        the model by adding the new model to the list of models and by
        making the new model the active model. If the list of models
        contains more than one model, it names the active model by
        combining the names of all the models. It then makes the buttons
        for saving the fit and for plotting the models visible.
        """
        model_class = self.model_combo.selected
        model_index = len(self.models)
        model_name = (model_class.replace('Model', '').replace(' ', '_')
                      + '_' + str(model_index+1))
        model = self.get_model_instance(model_class, model_name)
        try:
            if self.model:
                y = self.signal - self.model.eval(self.parameters, x=self.axis)
            else:
                y = self.signal
            parameters = model.guess(y, x=self.axis)
        except NotImplementedError:
            parameters = model.make_params()
        self.models.append({'name': model_name,
                            'class': model_class,
                            'model': model,
                            'parameters': parameters})
        self.add_model_parameters(model_index)
        self.write_parameters()
        if self.model is None:
            self.model = model
        else:
            self.model = self.model + model
        if len(self.models) > 1:
            self.composite_model += '+' + model_name
        else:
            self.composite_model = model_name
        self.set_button_visibility()

    def add_model_parameters(self, model_index):
        """
        Adds the parameters of a model to the model parameters box.

        Parameters
        ----------
        model_index : int
            The index of the model in the list of models.

        This method adds a row to the parameter grid for each parameter
        of the model, and adds a row to the remove combo box. If this is
        the first model, it adds the parameter, remove, adjust, and
        action layouts to the main layout, and adds the first model to
        the plot combo box.
        """
        self.add_model_rows(model_index)
        if self.first_time:
            self.layout.insertLayout(1, self.parameter_layout)
            self.layout.insertLayout(2, self.remove_layout)
            self.layout.insertLayout(3, self.adjust_layout)
            self.layout.insertLayout(4, self.action_layout)
            self.plot_combo.add('All')
            self.plot_combo.insertSeparator(1)
            self.plot_combo.insertSeparator(2)
            self.plot_combo.add('Composite Model')
            self.set_button_visibility()
        model_name = self.models[model_index]['name']
        self.remove_combo.add(self.expanded_name(model_name))
        self.remove_combo.select(self.expanded_name(model_name))
        self.plot_combo.insert(self.plot_combo.count()-2,
                               self.expanded_name(model_name))
        self.first_time = False

    def add_model_rows(self, model_index):
        """
        Adds a row to the parameter grid for each parameter of a model.

        Parameters
        ----------
        model_index : int
            The index of the model in the list of models.

        This method adds a row to the parameter grid for each parameter
        of the model, adds a row to the remove combo box, and adds the
        model to the plot combo box. If this is the first model, it adds
        the parameter, remove, adjust, and action layouts to the main
        layout.
        """
        model = self.models[model_index]['model']
        model_name = self.models[model_index]['name']
        parameters = self.models[model_index]['parameters']
        first_row = row = self.parameter_grid.rowCount()
        name = self.expanded_name(model_name)
        label_box = NXLabel(name)
        self.parameter_grid.addWidget(label_box, row, 0)
        for parameter in parameters:
            p = parameters[parameter]
            name = p.name.replace(model.prefix, '')
            if name == 'Fwhm':
                name = 'FWHM'
            p.box = {}
            p.box['value'] = NXLineEdit(align='right',
                                        slot=self.read_parameters)
            p.box['error'] = NXLabel()
            p.box['min'] = NXLineEdit('-inf', align='right')
            p.box['max'] = NXLineEdit('inf', align='right')
            p.box['fixed'] = NXCheckBox()
            p.box['expr'] = NXPushButton('Σ', self.edit_expression,
                                         checkable=True, width=50)
            self.parameter_grid.addWidget(NXLabel(name), row, 1)
            self.parameter_grid.addWidget(p.box['value'], row, 2)
            self.parameter_grid.addWidget(p.box['error'], row, 3)
            self.parameter_grid.addWidget(p.box['min'], row, 4)
            self.parameter_grid.addWidget(p.box['max'], row, 5)
            self.parameter_grid.addWidget(p.box['fixed'], row, 6,
                                          alignment=QtCore.Qt.AlignHCenter)
            self.parameter_grid.addWidget(p.box['expr'], row, 7)
            row += 1
        self.models[model_index]['row'] = first_row
        self.models[model_index]['label_box'] = label_box

    def remove_model(self):
        """
        Slot to be called when the user clicks the "Remove" button.

        This slot removes the selected model from the list of models,
        removes the parameters of the model from the model parameters
        box, and updates the model by subtracting the selected model from
        the list of models. If the list of models contains more than one
        model, it renames the active model by combining the names of all
        the models. If the list of models contains only one model, it
        makes the buttons for saving the fit and for plotting the models
        invisible. It then reads the parameters and makes the buttons
        for saving the fit and for plotting the models visible.
        """
        expanded_name = self.remove_combo.currentText()
        model_name = self.compressed_name(expanded_name)
        model_index = [self.models.index(m) for m in self.models
                       if m['name'] == model_name][0]
        parameters = self.models[model_index]['parameters']
        row = self.models[model_index]['row']
        for row in range(row, row+len(parameters)):
            for column in range(8):
                item = self.parameter_grid.itemAtPosition(row, column)
                if item is not None:
                    widget = item.widget()
                    if widget is not None:
                        widget.setVisible(False)
                        self.parameter_grid.removeWidget(widget)
                        widget.deleteLater()
        self.models.pop(model_index)
        self.plot_combo.removeItem(self.plot_combo.findText(expanded_name))
        self.remove_combo.removeItem(self.remove_combo.findText(expanded_name))
        self.model = None
        for i, m in enumerate(self.models):
            old_name = m['name']
            m['name'] = m['class'] + '_' + str(i+1)
            m['parameters'] = self.rename_parameters(m, old_name)
            m['label_box'].setText(self.expanded_name(m['name']))
            idx = self.parameter_grid.indexOf(m['label_box'])
            m['row'] = self.parameter_grid.getItemPosition(idx)[0]
            if i == 0:
                self.model = m['model']
                self.composite_model = m['name']
            else:
                self.model += m['model']
                self.composite_model += '+' + m['name']
            self.rename_model(old_name, m['name'])
        self.read_parameters()
        self.set_button_visibility()

    def rename_parameters(self, model, old_name):
        """
        Renames the parameters of a model.

        This method renames the parameters of a model by replacing the
        old name with the new name. It also updates the expressions of
        the parameters by replacing the old name with the new name. The
        _delay_asteval attribute of each parameter is set to True to
        prevent the expression from being evaluated until the model is
        evaluated. The method returns a new dictionary of parameters
        with the new names and the original parameter boxes.

        Parameters
        ----------
        model : dict
            The dictionary of the model.
        old_name : str
            The old name of the model.

        Returns
        -------
        parameters : dict
            The new dictionary of parameters with the new names and the
            original parameter boxes.
        """
        model['model'].prefix = model['name'] + '_'
        for p in model['parameters']:
            model['parameters'][p].name = model['parameters'][p].name.replace(
                old_name, model['name'])
            if model['parameters'][p].expr:
                model['parameters'][p].expr = (
                    model['parameters'][p].expr.replace(
                        old_name, model['name']))
            model['parameters'][p]._delay_asteval = True
        parameters = model['parameters'].copy()
        for p in parameters:
            old_p = p.replace(model['name'], old_name)
            parameters[p].box = model['parameters'][old_p].box
            parameters[p].box['error'].setText('')
        return parameters

    def rename_model(self, old_name, new_name):
        """
        Renames a model in the plot and remove combo boxes.

        This method takes the old name and new name of a model and
        renames the model in the plot combo box and the remove combo
        box. This method is used when the model is renamed in the
        Fit Panel.

        Parameters
        ----------
        old_name : str
            The old name of the model.
        new_name : str
            The new name of the model.
        """
        old_name, new_name = (self.expanded_name(old_name),
                              self.expanded_name(new_name))
        plot_index = self.plot_combo.findText(old_name)
        self.plot_combo.setItemText(plot_index, new_name)
        remove_index = self.remove_combo.findText(old_name)
        self.remove_combo.setItemText(remove_index, new_name)

    def compose_model(self):
        """
        Opens a dialog to allow the user to enter a composite model.

        When the method is called, it first closes any existing composite
        dialog. Then it creates a new instance of the CompositeDialog class
        and shows it. The dialog is the same as the one used to enter the
        initial model, but the window title is changed to "Composite Model".
        """
        if self.composite_dialog:
            try:
                self.composite_dialog.close()
            except Exception:
                pass
        self.composite_dialog = CompositeDialog(parent=self)
        self.composite_dialog.show()

    def eval_model(self, composite_text):
        """
        Evaluates a composite model.

        This method takes a composite model as a string and evaluates
        it. The string is a valid Python expression and can contain
        any valid Python syntax. The method first replaces all model
        names in the string with their corresponding model objects,
        and then evaluates the string using the built-in eval() function.
        If the string is invalid Python syntax, the method raises a
        NeXusError with the error message from the eval() function.

        Parameters
        ----------
        composite_text : str
            The composite model as a string.

        Returns
        -------
        model : lmfit.model.Model
            The evaluated composite model.

        Raises
        ------
        NeXusError
            If the string is invalid Python syntax.
        """
        models = {m['name']: m['model'] for m in self.models}
        text = composite_text
        for m in models:
            text = text.replace(m, f"models['{m}']")
        try:
            return eval(text)
        except Exception as error:
            raise NeXusError(str(error))

    def edit_expression(self):
        """
        Opens a dialog to edit the expression of a parameter.

        This method opens a dialog to edit the expression of a parameter
        when the 'expr' checkbox is checked. The dialog displays the
        current value of the parameter and allows the user to enter a
        mathematical expression. The expression can contain any valid
        Python syntax and can refer to any other parameter in the model
        by name. The dialog is closed when the user clicks the 'OK'
        button or presses Enter. The method then unchecks the 'expr'
        checkbox.
        """
        if self.expression_dialog:
            try:
                self.expression_dialog.close()
            except Exception:
                pass
        for m in self.models:
            for parameter in m['parameters']:
                p = m['parameters'][parameter]
                if p.box['expr'].isChecked():
                    self.expression_dialog = ExpressionDialog(p, parent=self)
                    self.expression_dialog.show()
                    p.box['expr'].setChecked(False)

    def eval_expression(self, parameter):
        """
        Evaluate the expression of a parameter.

        If the parameter has an expression, this method evaluates it
        using the current values of all the other parameters in the
        model. If the parameter does not have an expression, the
        method simply returns the current value of the parameter.

        Parameters
        ----------
        parameter : Parameter
            The parameter to evaluate.

        Returns
        -------
        value : float
            The value of the parameter.

        Raises
        ------
        NeXusError
            If the expression is invalid Python syntax.
        """
        try:
            if parameter.expr:
                return parameter._expr_eval(parameter.expr)
            else:
                return parameter.value
        except Exception as error:
            report_error(parameter.name, error)

    def read_parameters(self):
        """
        Read the values of all parameters in the model from the GUI.

        This method is called when the user clicks the 'OK' button in the
        model dialog. It reads the values of all parameters in the model
        from the GUI and updates the `value` attribute of each parameter
        object. If a parameter has an expression, this method also evaluates
        the expression using the current values of all the other parameters
        in the model and updates the `value` attribute of the parameter
        object with the result of the evaluation. The method returns a
        list of all the parameter objects in the model.

        Returns
        -------
        parameters : list
            A list of all the parameter objects in the model.
        """
        def make_float(value):
            try:
                return float(value)
            except Exception:
                return None
        for m in self.models:
            for parameter in m['parameters']:
                p = m['parameters'][parameter]
                p.value = make_float(p.box['value'].text())
                p.min = make_float(p.box['min'].text())
                p.max = make_float(p.box['max'].text())
                p.vary = not p.box['fixed'].isChecked()
        for m in self.models:
            for parameter in m['parameters']:
                p = m['parameters'][parameter]
                if p.expr:
                    p.value = self.eval_expression(p)
                    try:
                        p.box['value'].setText(format_float(p.value))
                    except Exception as error:
                        report_error(p.name, error)
                        return self.parameters
        return self.parameters

    def write_parameters(self):
        """
        Write the values of all parameters in the model to the GUI.

        This method is called when the user clicks the 'Cancel' button in the
        model dialog. It writes the values of all parameters in the model
        from the parameter objects to the GUI. If a parameter has an expression,
        this method also evaluates the expression using the current values
        of all the other parameters in the model and writes the result of
        the evaluation to the GUI.
        """
        def write_value(box, value, prefix=None):
            try:
                if prefix:
                    box.setText(prefix + ' ' + format_float(value))
                else:
                    box.setText(format_float(value))
            except TypeError:
                box.setText(' ')
        for m in self.models:
            for parameter in m['parameters']:
                p = m['parameters'][parameter]
                if p.expr:
                    write_value(p.box['value'], self.eval_expression(p))
                    p.box['fixed'].setCheckState(QtCore.Qt.Checked)
                    p.box['fixed'].setEnabled(False)
                else:
                    write_value(p.box['value'], p.value)
                if p.vary:
                    write_value(p.box['error'], p.stderr, prefix='+/-')
                    p.box['fixed'].setCheckState(QtCore.Qt.Unchecked)
                else:
                    p.box['fixed'].setCheckState(QtCore.Qt.Checked)
                write_value(p.box['min'], p.min)
                write_value(p.box['max'], p.max)

    def get_model(self, model=None, fit=False):
        """
        Return a NXdata object containing the model data.

        Parameters
        ----------
        model : Model, optional
            The model to evaluate. If not given, the model in the dialog
            is used.
        fit : bool, optional
            If True and the data have been fitted, the fitted parameters
            are used to evaluate the model. If False, the current
            parameters in the dialog are used to evaluate the model. If
            the data have not been fitted, the current parameters in the
            dialog are used to evaluate the model.

        Returns
        -------
        NXdata
            A NXdata object containing the model data.
        """
        if self.plot_checkbox.isChecked():
            x = self.axis
        else:
            xmin, xmax = self.get_limits()
            x = np.linspace(xmin, xmax, 1001)
        model_axis = NXfield(x, name='axis')
        if fit and self.fit:
            parameters = self.fit.params
        else:
            parameters = self.read_parameters()
        if model is None:
            model = self.model
        y = model.eval(parameters, x=x)
        if isinstance(y, float):
            y = y * np.ones(shape=x.shape)
        if fit:
            model_data = NXfield(y, name='fit')
        else:
            model_data = NXfield(y, name='model')
        return NXdata(model_data, model_axis, title=self.data.nxtitle)

    def get_limits(self):
        """
        Return the minimum and maximum values of the x-axis in the plot.

        Returns
        -------
        tuple
            A tuple of two floats, the minimum and maximum values of the
            x-axis.
        """
        return self.plotview.xtab.get_limits()

    @property
    def plot_min(self):
        """Return the minimum value of the x-axis in the plot."""
        return self.get_limits()[0]

    @property
    def plot_max(self):
        """Return the maximum value of the x-axis in the plot."""
        return self.get_limits()[1]

    def plot_data(self):
        """
        Plot the data in the current plotview.

        If the plotview is None, create a new plot of the data in the
        'Fit' plotview. Otherwise, change the color of the data in the
        current plotview to the color of the tab. Remove any other plots
        in the plotview. Set the linestyle cycle to the list of
        linestyles and plot the mask. Finally, raise the plotview to the
        top.
        """
        if self.fitview is None:
            self.fitview = NXPlotView('Fit')
            self.fitview.plot(self._data, fmt='o', color=self.color)
        elif self.fitview.label == 'Fit':
            self.fitview.plot(self._data,
                              xmin=self.plot_min, xmax=self.plot_max,
                              color=self.color)
        else:
            self.fitview.plots[self.data_num]['plot'].set_color(self.color)
        for label in ['label', 'legend_label']:
            self.fitview.plots[self.fitview.num][label] = self.data_label
        self.remove_plots()
        self.linestyle = cycle(self.linestyles)
        self.plot_mask()
        self.fitview.raise_()

    def plot_mask(self):
        """
        Plot the masked data in the current plotview.

        If there is no plotview, create a new plot of the data in the
        'Fit' plotview. Otherwise, remove any previous plots of the
        mask and plot the mask. Set the linestyle cycle to the list of
        linestyles and plot the mask. Finally, raise the plotview to the
        top.
        """
        if self.fitview is None:
            self.plot_data()
        mask_data = self.signal_mask()
        if mask_data:
            if self.mask_num in self.fitview.plots:
                self.fitview.plots[self.mask_num]['plot'].remove()
                del self.fitview.plots[self.mask_num]
            else:
                self.mask_num = self.next_plot_num()
            self.fitview.plot(mask_data, over=True, num=self.mask_num,
                              fmt='o', color='white', alpha=0.8)
            self.fitview.ytab.plotcombo.remove(self.mask_num)
            self.fitview.plots[self.mask_num]['legend_label'] = (
                f"{self.tab_label} Mask")
            self.fitview.plots[self.mask_num]['show_legend'] = False
            if self.fitview.plots[self.mask_num]['cursor']:
                self.fitview.plots[self.mask_num]['cursor'].remove()
            self.fitview.plots[self.mask_num]['cursor'] = None
            self.fitview.update_panels()
        self.remove_rectangle()

    def plot_model(self, model=False):
        """
        Plot the model in the current plotview.

        Parameters
        ----------
        model : bool or Model, optional
            If True, plot the composite model. If False, plot the model
            selected in the model combo box. If a Model object, plot the
            model. If the model is not the composite model, the model is
            plotted with a linestyle cycle of ['-', '--', '-.', ':'] and
            a color cycle of ['C0', 'C1', 'C2', ...]. If the model is the
            composite model and the data have been fitted, the model is
            plotted with a solid line. If the model is the composite model
            and the data have not been fitted, the model is plotted with a
            dashed line. The model is plotted with a legend label that
            includes the name of the tab and the name of the model.
        """
        if self.fitview is None:
            self.plot_data()
        model_name = self.plot_combo.currentText()
        if model is False:
            if model_name == 'Composite Model':
                if self.plot_dialog:
                    try:
                        self.plot_dialog.close()
                    except Exception:
                        pass
                self.plot_dialog = PlotModelDialog(parent=self)
                self.plot_dialog.show()
                return
            elif model_name != 'All':
                name = self.compressed_name(model_name)
                model = [m['model']
                         for m in self.models if m['name'] == name][0]
        num = self.next_plot_num()
        xmin, xmax = self.plot_min, self.plot_max
        if model_name == 'All':
            if self.fitted:
                fmt = '-'
            else:
                fmt = '--'
            self.fitview.plot(self.get_model(), fmt=fmt, color=self.color,
                              xmin=self.plot_min, xmax=self.plot_max,
                              over=True, num=num)
            if self.fitted:
                self.fitview.plots[num]['legend_label'] = (
                    f"{self.tab_label} Fit")
            else:
                self.fitview.plots[num]['legend_label'] = (
                    f"{self.tab_label} Model")
        else:
            self.fitview.plot(self.get_model(model), color=self.color,
                              marker=None, linestyle=next(self.linestyle),
                              xmin=self.plot_min, xmax=self.plot_max,
                              over=True, num=num)
            self.fitview.plots[num]['legend_label'] = (
                f"{self.tab_label} {model_name}")
        self.fitview.plots[num]['show_legend'] = False
        self.fitview.set_plot_limits(xmin=xmin, xmax=xmax)
        self.plot_nums.append(num)
        self.fitview.ytab.plotcombo.remove(num)
        self.fitview.ytab.plotcombo.select(self.data_num)
        self.remove_rectangle()
        self.fitview.update_panels()
        self.fitview.raise_()

    def next_plot_num(self):
        """
        Return the next available number for a plot.

        The next available number is defined as the first number greater
        than the data number that is not already in use as a plot number.
        The range of valid numbers is limited to 100 numbers, starting from
        1 plus 100 times the data number. If all valid numbers are in use,
        the next available number is the minimum of the valid numbers.

        Returns
        -------
        num : int
            The next available number for a plot.
        """
        min_num = self.data_num*100 + 1
        max_num = min_num + 98
        valid_nums = [n for n in self.fitview.plots if min_num <= n <= max_num]
        if valid_nums:
            return max(valid_nums) + 1
        else:
            return min_num

    def fit_data(self):
        """
        Fit the data with the current model and parameters.

        This function reads the current parameters from the GUI, creates
        a Model object, and fits the data using the Model object and the
        current parameters. If the data are weighted, the weights are
        passed to the Model object. If the fit is successful, the fit
        parameters are updated in the GUI and the fit results are
        reported in the GUI. If the fit fails, an error message is
        reported in the GUI.
        """
        self.read_parameters()
        if self.fit_checkbox.isChecked():
            weights = self.weights
        else:
            weights = None
        try:
            self.fit_status.setText('Fitting...')
            self.fit = self.model.fit(self.signal,
                                      params=self.parameters,
                                      weights=weights,
                                      x=self.axis,
                                      method=self.method,
                                      nan_policy='omit')
        except Exception as error:
            report_error("Fitting Data", error)
        if self.fit:
            if self.fit.success:
                self.fit_status.setText(
                    'Fit Successful Chi^2 = '
                    f'{format_float(self.fit.result.redchi)}')
            else:
                self.fit_status.setText(
                    'Fit Failed Chi^2 = '
                    f'{format_float(self.fit.result.redchi)}')
            self.parameters = self.fit.params
            self.write_parameters()
            self.set_button_visibility(fitted=True)
            self.fitted = True
        else:
            self.fit_status.setText('Fit failed')

    def report_fit(self):
        """Display the results of a fit in a message box."""
        if self.fit.result.errorbars:
            errors = 'Uncertainties estimated'
        else:
            errors = 'Uncertainties not estimated'
        text = (f'{self.fit.result.message}\n' +
                f'Chi^2 = {self.fit.result.chisqr}\n' +
                f'Reduced Chi^2 = {self.fit.result.redchi}\n' +
                f'{errors}\n' +
                f'No. of Function Evaluations = {self.fit.result.nfev}\n' +
                f'No. of Variables = {self.fit.result.nvarys}\n' +
                f'No. of Data Points = {self.fit.result.ndata}\n' +
                f'No. of Degrees of Freedom = {self.fit.result.nfree}\n' +
                f'{self.fit.fit_report()}')
        message_box = NXMessageBox('Fit Results', text, parent=self)
        message_box.setStandardButtons(QtWidgets.QMessageBox.Ok)
        spacer = QtWidgets.QSpacerItem(500, 0,
                                       QtWidgets.QSizePolicy.Minimum,
                                       QtWidgets.QSizePolicy.Expanding)
        layout = message_box.layout()
        layout.addItem(spacer, layout.rowCount(), 0, 1, layout.columnCount())
        message_box.exec()

    def store_fit(self):
        """Store the results of a fit in a NXprocess group."""
        if self.fit is None:
            self.fit_status.setText('Fit not available for saving')
            return
        self.read_parameters()
        group = NXprocess()
        group.attrs['composite_model'] = self.composite_model
        group['data'] = self.data
        for m in self.models:
            group[m['name']] = self.get_model(m['model'])
            parameters = NXparameters(attrs={'model': m['class']})
            for name in m['parameters']:
                p = self.fit.params[name]
                name = name.replace(m['model'].prefix, '')
                parameters[name] = NXfield(p.value, error=p.stderr,
                                           initial_value=p.init_value,
                                           min=str(p.min), max=str(p.max),
                                           vary=p.vary, expr=p.expr)
            group[m['name']].insert(parameters)
        group['program'] = 'lmfit'
        group['program'].attrs['version'] = lmfit_version
        group['title'] = 'Fit Results'
        group['fit'] = self.get_model(fit=True)
        fit = NXparameters()
        fit.nfev = self.fit.result.nfev
        fit.chisq = self.fit.result.chisqr
        fit.redchi = self.fit.result.redchi
        fit.message = self.fit.result.message
        group['statistics'] = fit
        group.note = NXnote(
            self.fit.result.message,
            f'Chi^2 = {self.fit.result.chisqr}\n'
            f'Reduced Chi^2 = {self.fit.result.redchi}\n'
            f'No. of Function Evaluations = {self.fit.result.nfev}\n'
            f'No. of Variables = {self.fit.result.nvarys}\n'
            f'No. of Data Points = {self.fit.result.ndata}\n'
            f'No. of Degrees of Freedom = {self.fit.result.nfree}\n'
            f'{self.fit.fit_report()}')
        return group

    def copy_fit(self):
        """
        Copy the results of a fit to the clipboard.

        The results of the fit are copied to the clipboard as a
        NXprocess group containing the model, data, and parameters. The
        group can then be pasted into another NeXus group in the tree.
        """
        group = self.store_fit()
        self.mainwindow.copied_node = self.mainwindow.copy_node(group)

    def save_fit(self):
        """Save the results of a fit to the scratch file.

        The results of the fit are written to the NeXpy scratch file as
        a NXprocess group containing the model, data, and parameters.
        """
        group = self.store_fit()
        self.write_group(group)

    def write_group(self, group):
        """
        Write a group to the NeXpy scratch file.

        The group is written to the first free number in the scratch file,
        starting from 'f1'.
        """
        if 'w0' not in self.tree:
            self.tree['w0'] = nxload(self.mainwindow.scratch_file, 'rw')
        ind = []
        for key in self.tree['w0']:
            try:
                if key.startswith('f'):
                    ind.append(int(key[1:]))
            except ValueError:
                pass
        if not ind:
            ind = [0]
        name = 'f'+str(sorted(ind)[-1]+1)
        self.tree['w0'][name] = group
        self.fit_status.setText(f'Parameters saved to w0/{name}')

    def copy_parameters(self):
        """Copy the current parameters to the clipboard.

        The parameters are copied to the clipboard as a NXprocess group
        containing the model, data, and parameters. The group can then
        be pasted into another NeXus group in the tree.
        """
        self.read_parameters()
        group = NXprocess()
        group.attrs['composite_model'] = self.composite_model
        group['data'] = self.data
        for m in self.models:
            group[m['name']] = self.get_model(m['model'])
            parameters = NXparameters(attrs={'model': m['class']})
            for n, p in m['parameters'].items():
                n = n.replace(m['model'].prefix, '')
                parameters[n] = NXfield(p.value, error=p.stderr,
                                        initial_value=p.init_value,
                                        min=str(p.min), max=str(p.max),
                                        vary=p.vary, expr=p.expr)
            group[m['name']].insert(parameters)
        group['title'] = 'Fit Model'
        group['model'] = self.get_model()
        self.mainwindow.copied_node = self.mainwindow.copy_node(group)

    def restore_parameters(self):
        """
        Restore the initial parameters of the fit model.

        This method is called when the 'Restore Initial Parameters'
        button is clicked. It resets the parameters to their initial
        values and updates the GUI.
        """
        self.parameters = self.fit.init_params
        self.write_parameters()
        self.fit_status.setText('Waiting to fit...')

    def on_button_release(self, event):
        """
        Handle a mouse button release event in the plot window.

        If the left button is released, remove any rectangle that was
        drawn. If the right button is released and a zoom is in
        progress, draw a rectangle with the zoom extent and make the
        mask button visible. In either case, the plot window is redrawn.
        """
        if self.fitview is None:
            return
        self.fitview.otab.release_zoom(event)
        if event.button == 1:
            self.remove_rectangle()
        elif event.button == 3 and self.fitview.zoom:
            self.xlo, self.xhi = self.fitview.zoom['x']
            self.ylo, self.yhi = self.fitview.zoom['y']
            self.draw_rectangle()
            self.clear_mask_button.setVisible(False)
            self.mask_button.setVisible(True)
        self.fitview.draw()

    def draw_rectangle(self):
        """
        Draw a rectangle in the plot window based on the zoom extent.

        This method is called when the right button is released in the
        plot window and a zoom is in progress. It sets the x and y
        coordinates of the rectangle to the zoom extent and makes the
        mask button visible. The plot window is then redrawn.
        """
        if self.fitview is None:
            return
        x, dx = self.xlo, self.xhi-self.xlo
        y, dy = self.ylo, self.yhi-self.ylo
        if self.rectangle:
            self.rectangle.set_bounds(x, y, dx, dy)
        else:
            self.rectangle = NXrectangle(x, y, dx, dy, plotview=self.fitview,
                                         facecolor='none',
                                         edgecolor=self.fitview._gridcolor)
            self.rectangle.set_linestyle('dashed')
            self.rectangle.set_linewidth(2)
        self.fitview.draw()

    def remove_rectangle(self):
        """
        Remove any rectangle that was drawn in the plot window.

        This method is called when the left button is released in the
        plot window. It removes any rectangle that was drawn and
        redraws the plot window.
        """
        if self.fitview is None:
            return
        if self.rectangle:
            self.rectangle.remove()
        self.rectangle = None
        self.fitview.draw()

    def remove_masks(self):
        """
        Remove the mask plot from the plot window.

        This method is called when the 'Clear Masks' button is clicked.
        It removes the mask plot from the plot window and redraws the
        plot window.
        """
        if self.fitview is None:
            return
        if self.mask_num in self.fitview.plots:
            self.fitview.plots[self.mask_num]['plot'].remove()
            del self.fitview.plots[self.mask_num]
            self.fitview.ytab.plotcombo.remove(self.mask_num)
        self.remove_rectangle()

    def remove_plots(self):
        """
        Remove any additional plots from the plot window.

        This method is called when the 'Fit Data' button is clicked.
        It removes any additional plots from the plot window and
        redraws the plot window.
        """
        if self.fitview is None:
            return
        for num in [n for n in self.plot_nums if n in self.fitview.plots]:
            self.fitview.plots[num]['plot'].remove()
            del self.fitview.plots[num]
        self.plot_nums = []
        if self.data_num in self.fitview.plots:
            self.fitview.num = self.data_num
            self.fitview.ytab.plotcombo.select(self.data_num)
        self.fitview.draw()
        self.fitview.update_panels()

    def apply(self):
        """
        Remove any additional plots from the plot window and replace
        them with a line plot of the fitted model. This method is called
        when the 'Apply' button is clicked. The plot window is then
        redrawn.
        """
        if self.fitview is None:
            return
        self.remove_plots()
        if self.model is not None:
            self.fitview.plot(self.get_model(), fmt='-', color=self.color,
                              over=True)

    def close(self):
        """
        Close this dialog and disconnect any signal connections.

        This method is called when the window is closed. It disconnects any
        signal connections and removes any masks and plots from the plot window.
        """
        if self.fitview is None:
            return
        self.fitview.canvas.mpl_disconnect(self.cid)
        self.remove_masks()
        self.remove_plots()


class CompositeDialog(NXDialog):

    def __init__(self, parent=None):

        """
        Initialize the dialog to edit a composite model.

        This dialog is initialized with the given parent and the
        composite model expression in the parent. The dialog contains
        a line edit box with the expression, a button to insert a model
        name, a combo box with the names of the models, and buttons to
        close the dialog. The dialog title is "Editing Composite Model".

        Parameters
        ----------
        parent : NXDialog
            The parent of the dialog.
        """
        super().__init__(parent=parent)

        self.parent = parent
        self.expression = NXLineEdit(self.parent.composite_model)
        self.add_model_button = NXPushButton('Insert Model', self.insert_model)
        self.model_combo = NXComboBox(items=[m['name']
                                             for m in self.parent.models])
        self.set_layout(self.expression,
                        self.make_layout(self.add_model_button,
                                         self.model_combo,
                                         'stretch',
                                         self.close_buttons(save=True)))
        self.set_title("Editing Composite Model")

    def insert_model(self):
        """
        Insert the selected model name into the expression box.

        This method is called when the user clicks the "Insert Model"
        button. It inserts the selected model name into the expression
        box at the current position of the text cursor.
        """
        self.expression.insert(self.model_combo.selected)

    def accept(self):
        """
        Accept the edited composite model expression.

        This method is called when the user clicks the "OK" button.
        It evaluates the expression in the expression box and
        assigns it to the model attribute of the parent. If the
        expression is invalid Python syntax, it raises a NeXusError
        with the error message from the eval() function.
        """
        try:
            self.parent.model = self.parent.eval_model(self.expression.text())
            self.parent.composite_model = self.expression.text()
            super().accept()
        except NeXusError as error:
            report_error("Editing Composite Model", error)


class PlotModelDialog(NXDialog):

    def __init__(self, parent=None):

        """
        Initialize the dialog to plot a composite model.

        Parameters
        ----------
        parent : NXDialog
            The parent of the dialog.

        The dialog contains a line edit box with the current composite
        model expression, a button to plot the model, and buttons to
        close the dialog. The dialog title is "Plotting Composite Model".
        """
        super().__init__(parent=parent)

        self.parent = parent
        self.expression = NXLineEdit(self.parent.composite_model)
        self.plot_model_button = NXPushButton('Plot Model', self.plot_model)
        self.set_layout(self.expression,
                        self.make_layout(self.plot_model_button,
                                         'stretch',
                                         self.close_buttons(close=True)))
        self.set_title("Plotting Composite Model")
        self.setMinimumWidth(400)

    def plot_model(self):
        """
        Plot the composite model with the current expression.

        This method is called when the user clicks the "Plot Model"
        button. It evaluates the expression in the expression box and
        passes the result to the plot_model method of the parent. If the
        expression is invalid Python syntax, it raises a NeXusError
        with the error message from the eval() function.
        """
        try:
            model = self.parent.eval_model(self.expression.text())
            self.parent.plot_model(model)
            super().accept()
        except NeXusError as error:
            report_error("Plotting Composite Model", error)


class ExpressionDialog(NXDialog):

    def __init__(self, parameter, parent=None):

        """
        Initialize the dialog to edit a parameter expression.

        The dialog is initialized with the given parameter and parent.
        The dialog contains a line edit box with the expression of the
        given parameter, a button to insert a parameter, a combo box
        with the parameter names, and buttons to close the dialog.
        The dialog title is in the form "Editing 'parameter_name' Expression".

        Parameters
        ----------
        parameter : NXParameter
            The parameter to be edited.
        parent : NXDialog, optional
            The parent of the dialog. The default is None.
        """
        super().__init__(parent=parent)

        self.parameter = parameter
        self.parent = parent
        self.expression = NXLineEdit(parameter.expr)
        self.add_parameter_button = NXPushButton('Insert Parameter',
                                                 self.insert_parameter)
        self.parameter_combo = NXComboBox(items=self.parent.parameters)
        self.set_layout(self.expression,
                        self.make_layout(self.add_parameter_button,
                                         self.parameter_combo,
                                         'stretch',
                                         self.close_buttons(save=True)))
        self.set_title(f"Editing '{parameter.name}' Expression")

    def insert_parameter(self):
        """
        Insert the selected parameter name into the expression box.

        This method is called when the user clicks the "Insert Parameter"
        button. It inserts the selected parameter name into the expression
        box at the current position of the text cursor.
        """
        self.expression.insert(self.parameter_combo.selected)

    def accept(self):
        """
        Accept the edited expression.

        This method is called when the user clicks the 'OK' button.
        It updates the parameter expression and evaluates the expression
        if it is not empty. If the expression is invalid Python syntax,
        it raises a NeXusError with the error message from the eval()
        function. The method calls the read_parameters() method of the
        parent to update the GUI and calls the accept() method of the
        parent to close the dialog.
        """
        try:
            p = self.parent.parameters[self.parameter.name]
            p.expr = self.expression.text()
            if p.expr:
                p.value = p._expr_eval(p.expr)
                p.box['value'].setText(format_float(p.value))
                p.box['fixed'].setChecked(True)
                p.box['fixed'].setEnabled(False)
            else:
                p.box['fixed'].setChecked(False)
                p.box['fixed'].setEnabled(True)
            self.parent.read_parameters()
            super().accept()
        except NeXusError as error:
            report_error("Editing Expression", error)



import numpy as np
from lmfit.model import Model


class PDFdecayModel(Model):
    r"""A model to describe the product of a decaying exponential and a
    Gaussian with three parameters: ``amplitude``, ``xi``, and ``sigma``

    .. math::

        f(x; A, \xi, \sigma) = A e^{[-{|x|}/\xi]} e^{[{-{x^2}/{{2\sigma}^2}}]}

    where the parameter ``amplitude`` corresponds to :math:`A`, ``xi`` to
    :math:`\xi`, and ``sigma`` to :math:`\sigma`.

    """

    def __init__(self, **kwargs):

        def pdfdecay(x, amplitude=1.0, xi=1.0, sigma=1.0):
            return amplitude * np.exp(-abs(x)/xi) * np.exp(-x**2/(2*sigma**2))

        super().__init__(pdfdecay, **kwargs)

    def guess(self, data, x=None, negative=False, **kwargs):
        """Estimate initial model parameter values from data."""
        sigma = np.sqrt(np.fabs((x**2*data).sum() / data.sum()))
        return self.make_params(amplitude=data.max(), xi=sigma, sigma=sigma)



import os

import pytest


def test_gui_import():
    os.environ['QT_API'] = 'pyqt5'
    try:
        import nexpy.gui.pyqt
        import nexpy.gui.consoleapp
        import nexpy.gui.dialogs
        import nexpy.gui.fitdialogs
        import nexpy.gui.importdialog
        import nexpy.gui.mainwindow
        import nexpy.gui.plotview
        import nexpy.gui.scripteditor
        import nexpy.gui.treeview
        import nexpy.gui.utils
        import nexpy.gui.widgets
    except ImportError as error:
        pytest.fail(str(error))


def test_reader_import():
    os.environ['QT_API'] = 'pyqt5'
    try:
        import nexpy.gui.pyqt
        import nexpy.readers.readspec
        import nexpy.readers.readstack
        import nexpy.readers.readtiff
        import nexpy.readers.readtxt
    except ImportError as error:
        pytest.fail(str(error))



# -----------------------------------------------------------------------------
# Copyright (c) 2025, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------

"""
Module to read a NumPy file and convert the array(s) to NeXus fields.
"""
from pathlib import Path
import numpy as np
from nexusformat.nexus import NeXusError, NXentry, NXfield

from nexpy.gui.importdialog import NXImportDialog


filetype = "NumPy Arrays"


class ImportDialog(NXImportDialog):

    """Dialog to import NumPy arrays"""

    def __init__(self, parent=None):

        super().__init__(parent=parent)
        self.set_layout(self.filebox(), self.buttonbox())
        self.set_title("Import " + str(filetype))

    def get_data(self):
        self.import_file = self.get_filename()
        if not self.import_file:
            raise NeXusError("No file specified")
        elif not Path(self.import_file).exists():
            raise NeXusError(f"File {self.import_file} does not exist")
        try:
            input = np.load(self.import_file)
            output = NXentry()
            if isinstance(input, np.ndarray):
                output['field'] = NXfield(input)
            elif isinstance(input, np.lib.npyio.NpzFile):
                output = NXentry()
                for name in input.files:
                    output[name] = NXfield(input[name])
            return output
        except Exception as error:
            raise NeXusError(f"Error reading {self.import_file}: {error}")



pv = NXPlotView('Chopper Plots')
phi = np.linspace(5., 95., 10)
chopper.entry.data[phi[0]:phi[0]+10].sum(0).plot(
    xmin=1900, xmax=2600, ymax=6000)
for i in range(10):
    (chopper.entry.data[phi[i]:phi[i]+10].sum(0)+500*i).oplot()



# -----------------------------------------------------------------------------
# Copyright (c) 2013-2025, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------

from pathlib import Path

from nexusformat.nexus import (NeXusError, NXdata, NXentry, NXfield, NXgroup,
                               NXlink, NXroot, nxload)

from .pyqt import QtCore, QtGui, QtWidgets
from .utils import (display_message, get_name, modification_time, report_error,
                    resource_icon)
from .widgets import NXSortModel


class NXtree(NXgroup):
    """
    NXtree group. This is a subclass of the NXgroup class.

    It is used as the invisible root item for NeXpy tree views.
    """
    _model = None
    _view = None
    _item = None
    _shell = {}
    _attrs = {}

    def __init__(self):
        """Initialize the NXtree group."""
        self._class = 'NXtree'
        self._name = 'tree'
        self._entries = {}

    def __setitem__(self, key, value):
        """
        Add a NeXus tree to the shell.

        Parameters
        ----------
        key : str
            The name of the tree in the shell.
        value : NXroot
            The root group of the tree.

        Raises
        ------
        NeXusError
            If the tree already exists in the shell.
        """
        if isinstance(value, NXroot):
            if key not in self._entries:
                value._group = self
                value._name = key
                self._entries[key] = value
                self._shell[key] = self._entries[key]
                value.set_changed()
            else:
                raise NeXusError("'"+key+"' already in the tree")
        else:
            raise NeXusError("Value must be an NXroot group")

    def __delitem__(self, key):
        """
        Delete a NeXus tree from the shell.

        Parameters
        ----------
        key : str
            The name of the tree in the shell.

        Raises
        ------
        NeXusError
            If the tree does not exist in the shell.
        """
        del self._entries[key]
        del self._shell[key]
        self.set_changed()

    def set_changed(self):
        """
        Mark the tree as changed.

        This method is called when the tree changes. It updates the tree
        view and the shell.
        """
        self.sync_shell_names()
        if self._model:
            self.sync_children(self._item)
            for row in range(self._item.rowCount()):
                for item in self._item.child(row).walk():
                    self.sync_children(item)
            index = self._item.index()
            if index.isValid():
                self._view.dataChanged(index, index)
            self._view.update()
            self._view.status_message(self._view.node)

    def sync_children(self, item):
        """
        Synchronize the children of a tree item with its NeXus group.

        If the NeXus group is a loaded NXgroup, this method adds
        children to the tree item if they are missing and removes them
        if they no longer exist in the NeXus group. This method should
        only be called by the set_changed method.

        Parameters
        ----------
        item : NXTreeItem
            The tree item to be synchronized.
        """
        if isinstance(item.node, NXgroup):
            children = []
            if item.hasChildren():
                for row in range(item.rowCount()):
                    children.append(item.child(row))
            names = [child.name for child in children]
            if item.node.entries_loaded:
                for name in item.node:
                    if name not in names:
                        item.appendRow(NXTreeItem(item.node[name]))
                for child in children:
                    if child.name not in item.node:
                        item.removeRow(child.row())
        item.node.set_unchanged()

    def add(self, node):
        """
        Add a NeXus group to the tree.

        If the NeXus group is a loaded NXgroup, it is added to the tree
        with its name. If the NeXus group is not loaded, it is added to
        a new NXroot group with a default name.

        Parameters
        ----------
        node : NXgroup
            The NeXus group to be added to the tree.

        Raises
        ------
        NeXusError
            If the node is not a NeXus group.
        """
        if isinstance(node, NXgroup):
            shell_names = self.get_shell_names(node)
            if shell_names:
                node.nxname = shell_names[0]
            if isinstance(node, NXroot):
                self[node.nxname] = node
                self[node.nxname]._file_modified = False
            elif isinstance(node, NXentry):
                group = NXroot(node)
                name = self.get_new_name()
                self[name] = group
                print(f"NeXpy: '{node.nxname}' added to tree in "
                      f"'{group.nxname}'")
            else:
                group = NXroot(NXentry(node))
                name = self.get_new_name()
                self[name] = group
                print(f"NeXpy: '{node.nxname}' added to tree in "
                      f"'{group.nxname}{node.nxgroup.nxpath}'")
        else:
            raise NeXusError("Only an NXgroup can be added to the tree")

    def load(self, filename, mode='r'):
        """
        Load a NeXus file into the tree.

        Parameters
        ----------
        filename : str
            The name of the NeXus file to be loaded.
        mode : str, optional
            The mode to open the file, by default 'r'.

        Returns
        -------
        NXroot
            The root group of the loaded tree.
        """
        name = self.get_name(filename)
        self[name] = nxload(filename, mode)
        return self[name]

    def reload(self, name):
        """
        Reload a NeXus tree.

        If the tree is a loaded NXroot, reload it from the file. If the
        tree is not a loaded NXroot, do nothing.

        Parameters
        ----------
        name : str
            The name of the tree to be reloaded.

        Returns
        -------
        NXroot
            The reloaded tree if it is a loaded NXroot, otherwise None.

        Raises
        ------
        NeXusError
            If the tree is not a loaded NXroot.
        """
        if name in self:
            if isinstance(self[name], NXroot):
                self[name].reload()
            return self[name]
        else:
            raise NeXusError(f"{name} not in the tree")

    def get_name(self, filename):
        """
        Return a name for a NeXus tree to be loaded.

        If a tree with the same filename exists, append a number to the
        name.

        Parameters
        ----------
        filename : str
            The name of the NeXus file to be loaded.

        Returns
        -------
        str
            A name for the loaded tree.
        """
        return get_name(filename, self._shell)

    def get_new_name(self):
        """
        Return a new name for a NeXus tree.

        The name is derived from the names of existing trees. If no
        trees exist, the name is 'w0'. If trees exist, the name is the
        largest number plus one, prefixed with 'w'.

        Returns
        -------
        str
            A new name for the tree.
        """
        ind = []
        for key in self._shell:
            try:
                if key.startswith('w'):
                    ind.append(int(key[1:]))
            except ValueError:
                pass
        if ind == []:
            ind = [0]
        return 'w'+str(sorted(ind)[-1]+1)

    def get_shell_names(self, node):
        """
        Return the names of the object in the shell namespace.

        Parameters
        ----------
        node : NXobject
            The NeXus object for which to retrieve the names.

        Returns
        -------
        list of str
            The names of the object in the shell namespace.
        """
        return [obj[0] for obj in self._shell.items() if id(obj[1]) == id(node)
                and not obj[0].startswith('_')]

    def sync_shell_names(self):
        """
        Ensure that the shell names are in sync with the tree.

        This method checks if the key in the tree is the same as the key
        in the shell namespace. If not, it assigns the tree key to the
        shell. If the previous key exists in the shell, it is deleted.

        """
        for key, value in self.items():
            shell_names = self.get_shell_names(value)
            if key not in shell_names:
                self._shell[key] = value
                if shell_names:
                    del self._shell[shell_names[0]]

    def node_from_file(self, fname):
        """
        Return the name of the tree item with filename fname.

        Parameters
        ----------
        fname : str
            The name of the file for which to return the tree item name.

        Returns
        -------
        str
            The name of the tree item if found, None otherwise.
        """
        fname = str(Path(fname).resolve())
        names = [name for name in self if self[name].nxfilename]
        try:
            return [name for name in names
                    if fname == self[name].nxfilename][0]
        except IndexError:
            return None


class NXTreeItem(QtGui.QStandardItem):

    """
    A subclass of the QtGui.QStandardItem class to return the data from
    an NXnode.
    """

    def __init__(self, node=None):
        """
        Constructor for the NXTreeItem class.

        Parameters
        ----------
        node : NXnode or None
            The NeXus node for which to create the tree item.

        Initializes the name, root, tree, and path attributes from the
        node. If the node is a link, sets the linked icon. If the node is
        a root, sets the locked, locked modified, unlocked, and unlocked
        modified icons. Calls the parent class constructor with the
        node name.
        """
        self.name = node.nxname
        self.root = node.nxroot
        self.tree = self.root.nxgroup
        self.path = self.root.nxname + node.nxpath
        if isinstance(node, NXlink):
            self._linked = resource_icon('link-icon.png')
        elif isinstance(node, NXroot):
            self._locked = resource_icon('lock-icon.png')
            self._locked_modified = resource_icon('lock-red-icon.png')
            self._unlocked = resource_icon('unlock-icon.png')
            self._unlocked_modified = resource_icon('unlock-red-icon.png')
        super().__init__(node.nxname)

    @property
    def node(self):
        """The selected node in the tree."""
        return self.tree[self.path]

    def __repr__(self):
        return f"NXTreeItem('{self.path}')"

    def text(self):
        """The name of the tree item."""
        return self.name

    def data(self, role=QtCore.Qt.DisplayRole):
        """
        Return the data for the tree item in the given role.

        Parameters
        ----------
        role : int, optional
            The role of the data, by default QtCore.Qt.DisplayRole

        Returns
        -------
        str or QIcon or None
            The name of the tree item for the DisplayRole and EditRole,
            a tooltip for the ToolTipRole, and a QIcon for the
            DecorationRole or None if the node is not a root or link.
        """
        if role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole:
            return self.name
        elif role == QtCore.Qt.ToolTipRole:
            try:
                tree = self.node.short_tree
                if tree.count('\n') > 50:
                    return '\n'.join(tree.split('\n')[0:50])+'\n...'
                else:
                    return tree
            except Exception:
                return ''
        elif role == QtCore.Qt.DecorationRole:
            try:
                if isinstance(self.node, NXroot):
                    if self.node.nxfilemode == 'r':
                        if self.node._file_modified:
                            return self._locked_modified
                        else:
                            return self._locked
                    elif self.node.nxfilemode == 'rw':
                        if self.node._file_modified:
                            return self._unlocked_modified
                        else:
                            return self._unlocked
                elif isinstance(self.node, NXlink):
                    return self._linked
                else:
                    return None
            except Exception:
                return None

    def children(self):
        """Return a list of child items."""
        items = []
        if self.hasChildren():
            for row in range(self.rowCount()):
                items.append(self.child(row))
        return items

    def walk(self):
        """Yield the tree item and its children."""
        yield self
        for child in self.children():
            for item in child.walk():
                yield item


class NXTreeView(QtWidgets.QTreeView):

    def __init__(self, tree, mainwindow):
        """
        Initialize a NeXus tree view.

        Parameters
        ----------
        tree : NXtree
            The root of the tree.
        mainwindow : NXMainWindow
            The main window of the application.
        parent : QWidget, optional
            The parent of the view.
        """
        super().__init__()

        self.tree = tree
        self.mainwindow = mainwindow
        self._model = QtGui.QStandardItemModel()
        self.proxymodel = NXSortModel(self)
        self.proxymodel.setSourceModel(self._model)
        self.proxymodel.setDynamicSortFilter(True)
        self.proxymodel.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)

        self.setModel(self.proxymodel)

        self._model.setColumnCount(1)
        self._model.setHorizontalHeaderItem(
            0, QtGui.QStandardItem('NeXus Data'))
        self.setSortingEnabled(True)
        self.sortByColumn(0, QtCore.Qt.AscendingOrder)

        self.setFocusPolicy(QtCore.Qt.ClickFocus)

        self.tree._item = self._model.invisibleRootItem()
        self.tree._item.node = self.tree
        self.tree._model = self._model
        self.tree._view = self
        self.tree._shell = self.mainwindow.user_ns

        self.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.setExpandsOnDoubleClick(False)
        self.selectionModel().selectionChanged.connect(self.selection_changed)
        self.expanded.connect(self.expand_node)

        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.check_modified_files)
        self.timer.start(1000)

        # Popup Menu
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.on_context_menu)

    def __repr__(self):
        return 'NXTreeView("nxtree")'

    def update(self):
        """Update the tree view."""
        super().update()

    def selection_changed(self):
        """
        Enable or disable menu items based on the selected node.

        This slot is connected to the selectionChanged signal of the
        tree view. It enables or disables actions based on whether the
        currently selected node is a root, group, link, field, or data
        item and whether it is modifiable or not.
        """
        self.mainwindow.savefile_action.setEnabled(False)
        self.mainwindow.duplicate_action.setEnabled(False)
        self.mainwindow.reload_action.setEnabled(False)
        self.mainwindow.remove_action.setEnabled(False)
        self.mainwindow.export_action.setEnabled(False)
        self.mainwindow.lockfile_action.setEnabled(False)
        self.mainwindow.unlockfile_action.setEnabled(False)
        self.mainwindow.backup_action.setEnabled(False)
        self.mainwindow.restore_backup_action.setEnabled(False)
        self.mainwindow.plot_data_action.setEnabled(False)
        self.mainwindow.plot_line_action.setEnabled(False)
        self.mainwindow.overplot_data_action.setEnabled(False)
        self.mainwindow.overplot_line_action.setEnabled(False)
        self.mainwindow.multiplot_data_action.setEnabled(False)
        self.mainwindow.multiplot_lines_action.setEnabled(False)
        self.mainwindow.plot_weighted_data_action.setEnabled(False)
        self.mainwindow.plot_image_action.setEnabled(False)
        self.mainwindow.view_action.setEnabled(False)
        self.mainwindow.validate_action.setEnabled(False)
        self.mainwindow.group_action.setEnabled(False)
        self.mainwindow.field_action.setEnabled(False)
        self.mainwindow.attribute_action.setEnabled(False)
        self.mainwindow.rename_action.setEnabled(False)
        self.mainwindow.delete_action.setEnabled(False)
        self.mainwindow.copydata_action.setEnabled(False)
        self.mainwindow.cutdata_action.setEnabled(False)
        self.mainwindow.pastedata_action.setEnabled(False)
        self.mainwindow.pastelink_action.setEnabled(False)
        self.mainwindow.link_action.setEnabled(False)
        self.mainwindow.signal_action.setEnabled(False)
        self.mainwindow.default_action.setEnabled(False)
        self.mainwindow.fit_action.setEnabled(False)
        self.mainwindow.fit_weighted_action.setEnabled(False)
        try:
            node = self.get_node()
        except Exception:
            node = None
        if node is None:
            self.mainwindow.reload_all_action.setEnabled(False)
            self.mainwindow.remove_all_action.setEnabled(False)
            self.mainwindow.collapse_action.setEnabled(False)
            return
        else:
            self.mainwindow.reload_all_action.setEnabled(True)
            self.mainwindow.remove_all_action.setEnabled(True)
            self.mainwindow.collapse_action.setEnabled(True)
            self.mainwindow.view_action.setEnabled(True)
            if node.nxgroup.is_modifiable():
                self.mainwindow.rename_action.setEnabled(True)
        if isinstance(node, NXroot):
            self.mainwindow.savefile_action.setEnabled(True)
            self.mainwindow.reload_action.setEnabled(True)
            self.mainwindow.remove_action.setEnabled(True)
            if node.nxfilemode:
                self.mainwindow.duplicate_action.setEnabled(True)
                if node.nxfilemode == 'r':
                    self.mainwindow.unlockfile_action.setEnabled(True)
                else:
                    self.mainwindow.lockfile_action.setEnabled(True)
                self.mainwindow.backup_action.setEnabled(True)
                if node.nxbackup:
                    self.mainwindow.restore_backup_action.setEnabled(True)
            if node.nxfilemode is None or node.nxfilemode == 'rw':
                if self.mainwindow.copied_node is not None:
                    self.mainwindow.pastedata_action.setEnabled(True)
                    self.mainwindow.pastelink_action.setEnabled(True)
            if node.nxfilemode is None:
                self.mainwindow.delete_action.setEnabled(True)
        else:
            self.mainwindow.copydata_action.setEnabled(True)
            if isinstance(node, NXgroup):
                self.mainwindow.validate_action.setEnabled(True)
            if isinstance(node, NXlink):
                self.mainwindow.link_action.setEnabled(True)
            if isinstance(node, NXdata):
                self.mainwindow.export_action.setEnabled(True)
            if node.is_modifiable():
                if isinstance(node, NXgroup):
                    self.mainwindow.group_action.setEnabled(True)
                    self.mainwindow.field_action.setEnabled(True)
                    if self.mainwindow.copied_node is not None:
                        self.mainwindow.pastedata_action.setEnabled(True)
                        self.mainwindow.pastelink_action.setEnabled(True)
                if not (isinstance(node, NXlink) or node.is_linked()):
                    self.mainwindow.attribute_action.setEnabled(True)
                self.mainwindow.cutdata_action.setEnabled(True)
                if not node.is_linked():
                    self.mainwindow.delete_action.setEnabled(True)
                if isinstance(node, NXentry) or isinstance(node, NXdata):
                    self.mainwindow.default_action.setEnabled(True)
                if isinstance(node, NXdata):
                    self.mainwindow.signal_action.setEnabled(True)
            try:
                if isinstance(node, NXdata) and node.plot_rank == 1:
                    self.mainwindow.fit_action.setEnabled(True)
                    if node.nxweights is not None:
                        self.mainwindow.fit_weighted_action.setEnabled(
                            True)
                elif (isinstance(node, NXgroup) and
                      ('fit' in node or 'model' in node)):
                    self.mainwindow.fit_action.setEnabled(True)
            except Exception:
                pass
        try:
            if (isinstance(node, NXdata) or 'default' in node.attrs or
                    node.is_numeric()):
                self.mainwindow.plot_data_action.setEnabled(True)
                if ((isinstance(node, NXgroup) and
                    node.nxsignal is not None and
                    node.nxsignal.plot_rank == 1) or
                        (isinstance(node, NXfield) and node.plot_rank == 1)):
                    self.mainwindow.plot_line_action.setEnabled(True)
                    if self.mainwindow.plotview.ndim == 1:
                        self.mainwindow.overplot_data_action.setEnabled(True)
                        self.mainwindow.overplot_line_action.setEnabled(True)
                    if 'auxiliary_signals' in node.attrs:
                        self.mainwindow.multiplot_data_action.setEnabled(True)
                        self.mainwindow.multiplot_lines_action.setEnabled(True)
                if (isinstance(node, NXgroup) and
                        node.plottable_data is not None):
                    if node.nxweights is not None:
                        self.mainwindow.plot_weighted_data_action.setEnabled(
                            True)
                    if (node.plottable_data.is_image() or
                            (isinstance(node, NXfield) and node.is_image())):
                        self.mainwindow.plot_image_action.setEnabled(True)
        except Exception:
            pass

    def expand_node(self, index):
        """
        Expand the node at index in the treeview.

        Parameters
        ----------
        index : QModelIndex
            The index of the node to expand.
        """
        item = self._model.itemFromIndex(self.proxymodel.mapToSource(index))
        if item and item.node:
            group = item.node
            for name in [n for n in group if isinstance(group[n], NXgroup)]:
                _entries = group[name].entries

    def addMenu(self, action):
        """Add an action to the menu."""
        if action.isEnabled():
            self.menu.addAction(action)

    def popMenu(self, node):
        """
        Create a context menu for the treeview.

        Parameters
        ----------
        node : Node
            The node at the context menu position.

        Returns
        -------
        menu : QMenu
            The context menu.
        """
        self.menu = QtWidgets.QMenu(self)
        self.addMenu(self.mainwindow.plot_data_action)
        self.addMenu(self.mainwindow.plot_line_action)
        self.addMenu(self.mainwindow.overplot_data_action)
        self.addMenu(self.mainwindow.overplot_line_action)
        self.addMenu(self.mainwindow.multiplot_data_action)
        self.addMenu(self.mainwindow.multiplot_lines_action)
        self.addMenu(self.mainwindow.plot_weighted_data_action)
        self.addMenu(self.mainwindow.plot_image_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.view_action)
        self.addMenu(self.mainwindow.validate_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.group_action)
        self.addMenu(self.mainwindow.field_action)
        self.addMenu(self.mainwindow.attribute_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.rename_action)
        self.addMenu(self.mainwindow.delete_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.copydata_action)
        self.addMenu(self.mainwindow.cutdata_action)
        self.addMenu(self.mainwindow.pastedata_action)
        self.addMenu(self.mainwindow.pastelink_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.link_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.fit_action)
        self.addMenu(self.mainwindow.fit_weighted_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.signal_action)
        self.addMenu(self.mainwindow.default_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.reload_action)
        self.addMenu(self.mainwindow.reload_all_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.remove_action)
        self.addMenu(self.mainwindow.remove_all_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.unlockfile_action)
        self.addMenu(self.mainwindow.lockfile_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.savefile_action)
        self.addMenu(self.mainwindow.duplicate_action)
        self.addMenu(self.mainwindow.export_action)
        self.addMenu(self.mainwindow.backup_action)
        self.addMenu(self.mainwindow.restore_backup_action)
        self.menu.addSeparator()
        self.addMenu(self.mainwindow.collapse_action)
        return self.menu

    def status_message(self, message):
        """
        Display a status message in the status bar of the main window.

        Parameters
        ----------
        message : str or Node
            The message to display. If a Node, it will be converted to a
            string using its _str_name() or _str_attrs() method.
        """
        if isinstance(message, NXfield) or isinstance(message, NXgroup):
            text = message._str_name()+' '+message._str_attrs()
        elif isinstance(message, NXlink):
            text = message._str_name()
        else:
            text = str(message)
        self.mainwindow.statusBar().showMessage(text.replace('\n', '; '))

    def check_modified_files(self):
        """
        Check the files in the tree for modifications.

        This checks the files in the tree for three conditions:

        1. If a file no longer exists, it is removed from the tree.
        2. If a file has been modified by another process, it is locked.
        3. If a file has been locked by another process, the lock flag is set.

        If any of these conditions are met, the tree is refreshed. If any of
        the checks fail, the refresh interval is set to 1 minute.
        """
        try:
            for key in list(self.tree._entries):
                node = self.tree._entries[key]
                if node.nxfilemode and not node.file_exists():
                    _dir = node.nxfile._filedir
                    if not Path(_dir).exists():
                        display_message("Files removed",
                                        f"'{_dir}' no longer exists")
                        for _key in [k for k in self.tree
                                     if self.tree[k].nxfile._filedir == _dir]:
                            del self.tree[_key]
                        break
                    else:
                        display_message(
                            "File removed",
                            f"'{node.nxfilename}' no longer exists")
                        del self.tree[key]
                elif node.is_modified():
                    node.lock()
                    node.nxfile.lock = True
                elif node.nxfilemode == 'rw':
                    nxfile = node.nxfile
                    if nxfile.is_locked() and nxfile.locked is False:
                        node.lock()
                        lock_time = modification_time(nxfile.lock_file)
                        display_message(f"'{node.nxname}' has been locked "
                                        "by an external process",
                                        f"Lock file created: {lock_time}")
                        nxfile.lock = True
            if self.timer.interval() > 1000:
                self.timer.setInterval(1000)
        except Exception as error:
            report_error("Checking Modified Files", error)
            self.timer.setInterval(60000)

    @property
    def node(self):
        """The currently selected node."""
        return self.get_node()

    def get_node(self):
        """Return the currently selected node using the proxy model."""
        item = self._model.itemFromIndex(
            self.proxymodel.mapToSource(self.currentIndex()))
        if item:
            return item.node
        else:
            return None

    def get_index(self, node):
        """
        Return the index of the node in the tree using the proxy model.

        Parameters
        ----------
        node : NXobject
            The NeXus node to find in the tree.

        Returns
        -------
        QModelIndex
            The index of the node in the tree using the proxy model.
        """
        items = self._model.findItems(node.nxname, QtCore.Qt.MatchRecursive)
        for item in items:
            if node is item.node:
                return self.proxymodel.mapFromSource(item.index())
        return None

    def select_node(self, node):
        """
        Select the node in the tree.

        Parameters
        ----------
        node : NXobject
            The NeXus node to select in the tree.
        """
        idx = self.get_index(node)
        if idx:
            self.setCurrentIndex(idx)
        self.selectionModel().select(self.currentIndex(),
                                     QtCore.QItemSelectionModel.Select)

    def select_top(self):
        """
        Select the first node in the tree and set focus to the treeview.

        This is called when the application is started and when a new
        file is opened. It is a convenience method to quickly select the
        first node in the tree and set focus to the treeview.
        """
        try:
            self.select_node(self.tree[self.tree.__dir__()[0]])
            self.setFocus()
        except Exception:
            pass

    def selectionChanged(self, new, old):
        """
        Called whenever the selection in the treeview changes.

        This method is connected to the selectionChanged signal of the
        treeview. It is called whenever the selection in the treeview
        changes. If an item is selected, the status bar of the main
        window is updated with the path of the selected node. If no
        item is selected, the status bar is cleared.

        Parameters
        ----------
        new : QItemSelection
            The new selection in the treeview.
        old : QItemSelection
            The old selection in the treeview.
        """
        super().selectionChanged(new, old)
        if new.indexes():
            node = self.get_node()
            self.status_message(node)
        else:
            self.status_message('')

    def collapse(self, index=None):
        """
        Collapse a node in the tree.

        Parameters
        ----------
        index : QModelIndex or None, optional
            The index of the node to collapse. If None, the entire tree
            is collapsed and the first node is selected. The default is
            None.
        """
        if index:
            super().collapse(index)
        else:
            self.collapseAll()
            self.setCurrentIndex(self.model().index(0, 0))

    def on_context_menu(self, point):
        """
        Called when the context menu is requested in the treeview.

        This method is connected to the customContextMenuRequested signal
        of the treeview. It is called whenever the context menu is
        requested in the treeview. If a node is selected, the context
        menu for that node is displayed at the requested point.

        Parameters
        ----------
        point : QPoint
            The position of the context menu in the treeview.
        """
        node = self.get_node()
        if node is not None:
            self.popMenu(self.get_node()).exec(self.mapToGlobal(point))

    def mouseDoubleClickEvent(self, event):
        """
        Called when the user double-clicks in the treeview.

        If a node is double-clicked, it is plotted in a new plot window
        if the Shift key is held down. Otherwise, it is plotted in the
        current default plot window.

        Parameters
        ----------
        event : QMouseEvent
            The double-click event in the treeview.
        """
        index = self.indexAt(event.pos())
        if index.isValid():
            if event.modifiers() & QtCore.Qt.ShiftModifier:
                self.mainwindow.plot_data(new_plot=True)
            else:
                self.mainwindow.plot_data()
        super().mouseDoubleClickEvent(event)


# -----------------------------------------------------------------------------
# Copyright (c) 2013-2021, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------
"""
Module to read in a SPEC file and convert it to NeXus.
"""
from importlib.util import find_spec
from pathlib import Path

import numpy as np
from nexusformat.nexus.tree import (NeXusError, NXdata, NXentry, NXfield,
                                    NXlog, NXroot)

from nexpy.gui.importdialog import NXImportDialog
from nexpy.gui.pyqt import QtWidgets, getOpenFileName
from nexpy.gui.widgets import NXLabel, NXLineEdit

filetype = "SPEC File"


class ImportDialog(NXImportDialog):
    """Dialog to import SPEC Scans."""

    def __init__(self, parent=None):

        super().__init__(parent=parent)

        if find_spec("spec2nexus") is None:
            raise NeXusError("Please install the 'spec2nexus' module")

        self.accepted = False
        self.import_file = None
        self.spec = None

        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setVisible(False)

        status_layout = QtWidgets.QHBoxLayout()
        status_layout.addWidget(self.progress_bar)
        status_layout.addStretch()
        status_layout.addWidget(self.close_buttons())

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(self.filebox())
        self.layout.addLayout(self.scanbox())
        self.layout.addLayout(status_layout)
        self.setLayout(self.layout)

        self.setWindowTitle("Import "+str(filetype))

    def scanbox(self):
        """Create widgets for specifying scan range to import."""
        scanminlabel = NXLabel("Min. Scan")
        self.scanmin = NXLineEdit(width=100, align='right')
        scanmaxlabel = NXLabel("Max. Scan")
        self.scanmax = NXLineEdit(width=100, align='right')

        scanbox = QtWidgets.QHBoxLayout()
        scanbox.addWidget(scanminlabel)
        scanbox.addWidget(self.scanmin)
        scanbox.addWidget(scanmaxlabel)
        scanbox.addWidget(self.scanmax)
        return scanbox

    def get_scan_numbers(self):
        return sorted([int(s) for s in self.spec.getScanNumbers()])

    def choose_file(self):
        """Opens file dialog, set file text box to the chosen path."""
        from spec2nexus.spec import SpecDataFile
        dirname = self.get_default_directory(self.filename.text())
        filename = getOpenFileName(self, 'Open file', dirname)
        if Path(filename).exists():
            self.filename.setText(str(filename))
            self.spec = SpecDataFile(self.get_filename())
            self.set_default_directory(Path(filename).parent)
            all_scans = self.get_scan_numbers()
            scan_min = all_scans[0]
            self.scanmin.setText(str(scan_min))
            scan_max = all_scans[-1]
            self.scanmax.setText(str(scan_max))

    def get_data(self):
        """Read the data and return :class:`NXroot` or :class:`NXentry`."""
        self.import_file = self.get_filename()
        if not Path(self.import_file).exists():
            return None
        if self.spec is None:
            return None
        scan_min = int(self.scanmin.text())
        scan_max = int(self.scanmax.text())
        all_scans = self.get_scan_numbers()
        scans = [s for s in all_scans if scan_min <= s <= scan_max]
        self.spec.progress_bar = self.progress_bar
        self.spec.update_progress = self.update_progress
        return Parser(self.spec).toTree(scans)


class Parser:
    """Parse the spec data file object."""

    def __init__(self, spec_data=None):
        """Instance of :class:`spec2nexus.prjPySpec.SpecDataFile`"""
        self.SPECfile = spec_data
        self.progress_bar = spec_data.progress_bar
        self.update_progress = spec_data.update_progress

    def openFile(self, filename):
        """Open the SPEC file and get its data."""
        from spec2nexus.spec import SpecDataFile
        if Path(filename).exists():
            self.SPECfile = SpecDataFile(filename)

    def toTree(self, scan_list=[]):
        """Convert scans from SPEC file into NXroot object and structure.

        Called from nexpy.readers.readspec.ImportDialog.get_data__prjPySpec()
        after clicking <Ok> in dialog.

        Each scan in the range from self.scanmin to self.scanmax (inclusive)
        will be converted to a NXentry.  Scan data will go in a NXdata where
        the signal=1 is the last column and the corresponding axes= is the
        first column.

        :param [int] scanlist
        :raises: ValueError is Min or Max scan number are not given properly
        """
        import spec2nexus
        from spec2nexus import utils

        # check that scan_list is valid
        if len(scan_list) == 0:
            return None

        if self.SPECfile is None:
            return None

        complete_scan_list = list(self.SPECfile.scans)
        for key in [str(s) for s in scan_list]:
            if key not in complete_scan_list:
                msg = 'scan ' + str(key) + ' was not found'
                raise ValueError(msg)

        root = NXroot()

        root.attrs['spec2nexus'] = str(spec2nexus.__version__)
        header0 = self.SPECfile.headers[0]
        root.attrs['SPEC_file'] = self.SPECfile.fileName
        root.attrs['SPEC_epoch'] = header0.epoch
        root.attrs['SPEC_date'] = utils.iso8601(header0.date)
        root.attrs['SPEC_comments'] = '\n'.join(header0.comments)
        try:
            c = header0.comments[0]
            user = c[c.find('User = '):].split('=')[1].strip()
            root.attrs['SPEC_user'] = user
        except Exception:
            pass
        root.attrs['SPEC_num_headers'] = len(self.SPECfile.headers)

        self.progress_bar.setVisible(True)
        self.progress_bar.setRange(scan_list[0], scan_list[-1])
        for key in [str(s) for s in scan_list]:
            scan = self.SPECfile.getScan(key)
            scan.interpret()
            entry = NXentry()
            entry.title = str(scan)
            entry.date = utils.iso8601(scan.date)
            entry.command = scan.scanCmd
            entry.scan_number = NXfield(scan.scanNum)
            entry.comments = '\n'.join(scan.comments)
            entry.data = self.scan_NXdata(
                scan)            # store the scan data
            entry.positioners = self.metadata_NXlog(
                scan.positioner, 'SPEC positioners (#P & #O lines)')
            if hasattr(scan, 'metadata') and len(scan.metadata) > 0:
                entry.metadata = self.metadata_NXlog(
                    scan.metadata,
                    'SPEC metadata (UNICAT-style #H & #V lines)')

            if len(scan.G) > 0:
                entry.G = NXlog()
                desc = "SPEC geometry arrays, defined by SPEC diffractometer"
                # e.g.: SPECD/four.mac
                # http://certif.com/spec_manual/fourc_4_9.html
                entry.G.attrs['description'] = desc
                for item, value in scan.G.items():
                    entry.G[item] = NXfield(list(map(float, value.split())))
            if scan.T != '':
                entry['counting_basis'] = NXfield(
                    'SPEC scan with constant counting time')
                entry['T'] = NXfield(float(scan.T))
                entry['T'].units = 's'
                entry['T'].description = 'Scan with constant counting time'
            elif scan.M != '':
                entry['counting_basis'] = NXfield(
                    'SPEC scan with constant monitor count')
                entry['M'] = NXfield(float(scan.M))
                entry['M'].units = 'counts'
                entry['M'].description = 'Scan with constant monitor count'
            if scan.Q != '':
                entry['Q'] = NXfield(list(map(float, scan.Q)))
                entry['Q'].description = 'hkl at start of scan'

            root['scan_' + str(key)] = entry

            self.progress_bar.setValue(int(key))
            self.update_progress()

        return root

    def scan_NXdata(self, scan):
        """Return the scan data in an NXdata object."""

        nxdata = NXdata()

        if len(scan.data) == 0:       # what if no data?
            # since no data available, provide trivial, fake data
            # keeping the NXdata base class compliant with the NeXus standard
            nxdata.attrs['description'] = 'SPEC scan has no data'
            nxdata['noSpecData_y'] = NXfield([0, 0])   # primary Y axis
            nxdata['noSpecData_x'] = NXfield([0, 0])   # primary X axis
            nxdata.nxsignal = nxdata['noSpecData_y']
            nxdata.nxaxes = [nxdata['noSpecData_x'], ]
            return nxdata

        nxdata.attrs['description'] = 'SPEC scan data'

        scan_type = scan.scanCmd.split()[0]
        if scan_type in ('mesh', 'hklmesh'):
            # hklmesh  H 1.9 2.1 100  K 1.9 2.1 100  -800000
            self.parser_mesh(nxdata, scan)
        elif scan_type in ('hscan', 'kscan', 'lscan', 'hklscan'):
            # hklscan  1.00133 1.00133  1.00133 1.00133  2.85 3.05  200 -400000
            h_0, h_N, k_0, k_N, l_0, l_N = scan.scanCmd.split()[1:7]
            if h_0 != h_N:
                axis = 'H'
            elif k_0 != k_N:
                axis = 'K'
            elif l_0 != l_N:
                axis = 'L'
            else:
                axis = 'H'
            self.parser_1D_columns(nxdata, scan)
            nxdata.nxaxes = nxdata[axis]
        else:
            self.parser_1D_columns(nxdata, scan)

        return nxdata

    def parser_1D_columns(self, nxdata, scan):
        """Generic data parser for 1-D column data."""
        from spec2nexus import utils
        for column in scan.L:
            if column in scan.data:
                clean_name = utils.sanitize_name(nxdata, column)
                nxdata[clean_name] = NXfield(scan.data[column])
                nxdata[clean_name].original_name = column

        signal = utils.sanitize_name(
            nxdata, scan.column_last)  # primary Y axis
        axis = utils.sanitize_name(
            nxdata, scan.column_first)  # primary X axis
        nxdata.nxsignal = nxdata[signal]
        nxdata.nxaxes = nxdata[axis]

        self.parser_mca_spectra(nxdata, scan, axis)

    def parser_mca_spectra(self, nxdata, scan, primary_axis_label):
        """Parse for optional MCA spectra."""
        if '_mca_' in scan.data:        # check for it
            for mca_key, mca_data in scan.data['_mca_'].items():
                key = "__" + mca_key
                nxdata[key] = NXfield(mca_data)
                nxdata[key].units = "counts"
                ch_key = key + "_channel"
                nxdata[ch_key] = NXfield(range(1, len(mca_data[0])+1))
                nxdata[ch_key].units = 'channel'
                axes = (primary_axis_label, ch_key)
                nxdata[key].axes = ':'.join(axes)

    def parser_mesh(self, nxdata, scan):
        """Data parser for 2-D mesh and hklmesh."""
        # 2-D parser: http://www.certif.com/spec_help/mesh.html
        # mesh motor1 start1 end1 intervals1 motor2 start2 end2 intervals2 time
        # 2-D parser: http://www.certif.com/spec_help/hklmesh.html
        #  hklmesh Q1 start1 end1 intervals1 Q2 start2 end2 intervals2 time
        # mesh:    nexpy/examples/33id_spec.dat  scan 22  (MCA gives 3-D data)
        # hklmesh: nexpy/examples/33bm_spec.dat  scan 17  (no MCA data)
        from spec2nexus import utils
        (label1, start1, end1, intervals1, label2, start2, end2,
         intervals2, time) = scan.scanCmd.split()[1:]
        if label1 not in scan.data:
            label1 = scan.L[0]      # mnemonic v. name
        if label2 not in scan.data:
            label2 = scan.L[1]      # mnemonic v. name
        axis1 = scan.data.get(label1)
        axis2 = scan.data.get(label2)
        intervals1, intervals2 = int(intervals1), int(intervals2)
        start1, end1 = float(start1), float(end1)
        start2, end2 = float(start2), float(end2)
        time = float(time)
        if len(axis1) < intervals1:  # stopped scan before second row started
            self.parser_1D_columns(nxdata, scan)        # fallback support
            # TODO: what about the MCA data in this case?
        else:
            axis1 = axis1[0:intervals1+1]
            axis2 = [axis2[row]
                     for row in range(len(axis2)) if row % (intervals1+1) == 0]

            column_labels = scan.L
            column_labels.remove(label1)    # special handling
            column_labels.remove(label2)    # special handling
            if scan.scanCmd.startswith('hkl'):
                # find the reciprocal space axis held constant
                label3 = [
                    key for key in ('H', 'K', 'L')
                    if key not in (label1, label2)][0]
                axis3 = scan.data.get(label3)[0]
                nxdata[label3] = NXfield(axis3)
                column_labels.remove(label3)    # already handled

            nxdata[label1] = NXfield(axis1)    # 1-D array
            nxdata[label2] = NXfield(axis2)    # 1-D array

            # build 2-D data objects
            data_shape = [len(axis2), len(axis1)]
            for label in column_labels:
                axis = np.array(scan.data.get(label))
                clean_name = utils.sanitize_name(nxdata, label)
                nxdata[clean_name] = NXfield(
                    utils.reshape_data(axis, data_shape))
                nxdata[clean_name].original_name = label

            signal_axis_label = utils.sanitize_name(nxdata, scan.column_last)
            nxdata.nxsignal = nxdata[signal_axis_label]
            nxdata.nxaxes = [nxdata[label2], nxdata[label1]]

        if '_mca_' in scan.data:    # 3-D array
            # TODO: ?merge with parser_mca_spectra()?
            for mca_key, mca_data in scan.data['_mca_'].items():
                key = "__" + mca_key

                spectra_lengths = list(map(len, mca_data))
                num_channels = max(spectra_lengths)
                if num_channels != min(spectra_lengths):
                    msg = 'MCA spectra have different lengths'
                    msg += ' in scan #' + str(scan.scanNum)
                    msg += ' in file ' + str(scan.specFile)
                    raise ValueError(msg)

                data_shape += [num_channels, ]
                mca = np.array(mca_data)
                nxdata[key] = NXfield(utils.reshape_data(mca, data_shape))
                nxdata[key].units = "counts"

                try:
                    # use MCA channel numbers as known at time of scan
                    chan1 = scan.MCA['first_saved']
                    chanN = scan.MCA['last_saved']
                    channel_range = range(chan1, chanN+1)
                except Exception:
                    # basic indices
                    channel_range = range(1, num_channels+1)

                ch_key = key + "_channel"
                nxdata[ch_key] = NXfield(channel_range)
                nxdata[ch_key].units = 'channel'
                axes = (label1, label2, ch_key)
                nxdata[key].axes = ':'.join(axes)

    def metadata_NXlog(self, spec_metadata, description):
        """Return the specific metadata in an NXlog object."""
        from spec2nexus import utils
        nxlog = NXlog()
        nxlog.attrs['description'] = description
        for subkey, value in spec_metadata.items():
            clean_name = utils.sanitize_name(nxlog, subkey)
            nxlog[clean_name] = NXfield(value)
            nxlog[clean_name].original_name = subkey
        return nxlog



# -----------------------------------------------------------------------------
# Copyright (c) 2013-2021, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------

"""
Module to read in a TIFF file using 'tifffile' and convert it to NeXus.
"""
import numpy as np
from nexpy.gui.importdialog import NXImportDialog
from nexusformat.nexus import NeXusError, NXdata, NXentry, NXfield

filetype = "TIFF Image"


class ImportDialog(NXImportDialog):
    """Dialog to import a TIFF image"""

    def __init__(self, parent=None):

        super().__init__(parent=parent)

        self.set_layout(self.filebox(), self.buttonbox())

        self.set_title("Import "+str(filetype))

    def get_data(self):
        self.import_file = self.get_filename()
        if not self.import_file:
            raise NeXusError("No file specified")
        try:
            import fabio
        except ImportError:
            raise NeXusError("Please install the 'fabio' module")
        im = fabio.open(str(self.import_file))
        z = NXfield(im.data, name='z')
        y = NXfield(np.arange(z.shape[0], dtype=float), name='y')
        x = NXfield(np.arange(z.shape[1], dtype=float), name='x')

        return NXentry(NXdata(z, (y, x)))



# -----------------------------------------------------------------------------
# Copyright (c) 2025, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------
import warnings

from .dialogs import *

warnings.warn(
    "The module 'nexpy.gui.datadialogs' is deprecated and will be removed " +
    "in a future release. Please use 'nexpy.gui.dialogs' instead.",
    DeprecationWarning,
    stacklevel=2
)
